# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

from PyQt4.QtCore import QObject
from PyQt4.QtCore import SIGNAL
from PyQt4.QtCore import pyqtSignal

from ninja_ide.core import settings
from ninja_ide.core import file_manager
from ninja_ide.core import plugin_util
from ninja_ide.gui.main_panel import itab_item
from ninja_ide.gui.main_panel import main_container
from ninja_ide.gui import actions
from ninja_ide.gui.explorer import explorer_container


###############################################################################
# PLUGINS SERVICES
###############################################################################


class MainService(QObject):
    """
    Main Interact whith NINJA-IDE
    """
    # SIGNALS
    editorKeyPressEvent = pyqtSignal("QEvent")
    beforeFileSaved = pyqtSignal("QString")
    fileSaved = pyqtSignal("QString")
    currentTabChanged = pyqtSignal("QString")
    fileExecuted = pyqtSignal("QString")
    fileOpened = pyqtSignal("QString")

    def __init__(self):
        QObject.__init__(self)
        self._main = main_container.MainContainer()
        self._action = actions.Actions()
        self._explorer = explorer_container.ExplorerContainer()
        #Connect signals
        self.connect(self._main, SIGNAL("editorKeyPressEvent(QEvent)"),
                     self._keyPressEvent)
        self.connect(self._main, SIGNAL("beforeFileSaved(QString)"),
                     self._beforeFileSaved)
        self.connect(self._main, SIGNAL("fileSaved(QString)"),
                     self._fileSaved)
        self.connect(self._main, SIGNAL("currentTabChanged(QString)"),
                     self._currentTabChanged)
        self.connect(self._action, SIGNAL("fileExecuted(QString)"),
                     self._fileExecuted)
        self.connect(self._main, SIGNAL("fileOpened(QString)"),
                     self._fileOpened)

###############################################################################
# Get main GUI Objects
###############################################################################

    def get_tab_manager(self):
        """
        Returns the TabWidget (ninja_ide.gui.main_panel.tab_widget.TabWidget)
        subclass of QTabWidget
        """
        return self._main.actualTab

###############################################################################
# END main GUI Objects
###############################################################################

    def add_menu(self, menu, lang=".py"):
        """
        Add an *extra context menu* to the editor context menu
        """
        itab_item.ITabItem.add_extra_menu(menu, lang=lang)

    def get_opened_documents(self):
        """
        Returns the opened documents
        """
        documents_data = self._main.get_opened_documents()
        #returns ONLY names!
        return [doc_data[0] for doc_list in documents_data
                for doc_data in doc_list]

    def get_project_owner(self, editorWidget=None):
        """
        Return the project where this file belongs, or an empty string.
        """
        #if not editor try to get the current
        if editorWidget is None:
            editorWidget = self._main.get_actual_editor()
        belongs = ''
        if editorWidget is None:
            return belongs
        #get the opened projects
        opened_projects_obj = self._explorer.get_opened_projects()
        for project in opened_projects_obj:
            if file_manager.belongs_to_folder(project.path, editorWidget.ID):
                belongs = project.path
                break
        return belongs

    def get_file_syntax(self, editorWidget=None):
        """Return the syntax for this file -> {}."""
        if editorWidget is None:
            editorWidget = self._main.get_actual_editor()

        if editorWidget is not None:
            ext = file_manager.get_file_extension(editorWidget.ID)
            lang = settings.EXTENSIONS.get(ext, '')
            syntax = settings.SYNTAX.get(lang, {})
            return syntax
        return {}

    def add_editor(self, fileName="", content=None, syntax=None):
        """
        Create a new editor
        """
        editor = self._main.add_editor(fileName=fileName, syntax=syntax)
        if content:
            editor.setPlainText(content)
        return editor

    def get_editor(self):
        """
        Returns the actual editor (instance of ninja_ide.gui.editor.Editor)
        This method could return None
        """
        return self._main.get_actual_editor()

    def get_editor_path(self):
        """
        Returns the actual editor path
        This method could return None if there is not an editor
        """
        editor = self._main.get_actual_editor()
        if editor:
            return editor.ID
        return None

    def get_editor_encoding(self, editorWidget=None):
        """
        Returns the editor encoding
        """
        if editorWidget is None:
            editorWidget = self._main.get_actual_editor()

        if editorWidget is not None:
            return editorWidget.encoding
        return None

    def get_text(self):
        """
        Returns the plain text of the current editor
        or None if thre is not an editor.
        """
        editor = self._main.get_actual_editor()
        if editor:
            return editor.get_text()
        return

    def get_selected_text(self):
        """
        Returns the selected text of and editor.
        This method could return None
        """
        editor = self._main.get_actual_editor()
        if editor:
            return editor.textCursor().selectedText()
        return None

    def insert_text(self, text):
        """
        Insert text in the current cursor position
        @text: string
        """
        editor = self._main.get_actual_editor()
        if editor:
            editor.insertPlainText(text)

    def jump_to_line(self, lineno):
        """
        Jump to a specific line in the current editor
        """
        self._main.editor_jump_to_line(lineno=lineno)

    def get_lines_count(self):
        """
        Returns the count of lines in the current editor
        """
        editor = self._main.get_actual_editor()
        if editor:
            return editor.get_lines_count()
        return None

    def save_file(self):
        """
        Save the actual file
        """
        self._main.save_file()

    def open_files(self, files, mainTab=True):
        """
        Open many files
        """
        self._main.open_files(self, files, mainTab=mainTab)

    def open_file(self, fileName='', cursorPosition=0,
                  positionIsLineNumber=False):
        """
        Open a single file, if the file is already open it get focus
        """
        self._main.open_file(filename=fileName, cursorPosition=cursorPosition,
                             positionIsLineNumber=positionIsLineNumber)

    def open_image(self, filename):
        """
        Open a single image
        """
        self._main.open_image(filename)

    def get_actual_tab(self):
        """
        Returns the actual widget
        """
        return self._main.get_actual_widget()

    # SIGNALS
    def _keyPressEvent(self, qEvent):
        """
        Emit the signal when a key is pressed
        @event: QEvent
        """
        self.editorKeyPressEvent.emit(qEvent)

    def _beforeFileSaved(self, fileName):
        """
        Signal emitted before save a file
        """
        self.beforeFileSaved.emit(fileName)

    def _fileSaved(self, fileName):
        """
        Signal emitted after save a file
        """
        fileName = fileName.split(":")[-1].strip()
        self.fileSaved.emit(fileName)

    def _currentTabChanged(self, fileName):
        """
        Signal emitted when the current tab changes
        """
        self.currentTabChanged.emit(fileName)

    def _fileExecuted(self, fileName):
        """
        Signal emitted when the file is executed
        """
        self.fileExecuted.emit(fileName)

    def _fileOpened(self, fileName):
        """
        Signal emitted when the file is opened
        """
        self.fileOpened.emit(fileName)


class ToolbarService(QObject):
    """
    Interact with the Toolbar
    """
    def __init__(self, toolbar):
        QObject.__init__(self)
        self._toolbar = toolbar

    def add_action(self, action):
        """
        Add an action to the Toolbar
        @action: Should be an instance(or subclass) of QAction
        """
        settings.add_toolbar_item_for_plugins(action)
        self._toolbar.addAction(action)


class MenuAppService(QObject):
    """
    Interact with the Plugins Menu
    """
    def __init__(self, plugins_menu):
        QObject.__init__(self)
        self._plugins_menu = plugins_menu

    def add_menu(self, menu):
        """
        Add an extra menu to the Plugin Menu of NINJA
        """
        self._plugins_menu.addMenu(menu)

    def add_action(self, action):
        """
        Add an action to the Plugin Menu of NINJA
        """
        self._plugins_menu.addAction(action)


#class ProjectTypeService(QObject):
#    """
#    Interact with the New Project Wizard
#    """
#    def __init__(self):
#        QObject.__init__(self)
#
#    def set_project_type_handler(self, project_type, project_type_handler):
#        """
#        Add a new Project Type and the handler for it
#        example:
#            foo_project_handler = FooProjectHandler(...)
#            set_project_type_handler('Foo Project', foo_project_handler)
#        Then 'Foo Project' will appear in the New Project wizard
#            and foo_project_handler instance controls the wizard
#
#        Note: project_type_handler SHOULD have a special interface see
#        ninja_ide.core.plugin_interfaces
#        """
#        settings.set_project_type_handler(project_type, project_type_handler)
#
#
#class TreeSymbolsService(QObject):
#    """
#    Interact with the symbols tree
#    """
#    def __init__(self):
#        QObject.__init__(self)
#
#    def set_symbols_handler(self, file_extension, symbols_handler):
#        """
#        Add a new Symbol's handler for the given file extension
#        example:
#            cpp_symbols_handler = CppSymbolHandler(...)
#            set_symbols_handler('cpp', cpp_symbols_handler)
#        Then all symbols in .cpp files will be handle by cpp_symbols_handler
#
#        Note: symbols_handler SHOULD have a special interface see
#        ninja_ide.core.plugin_interfaces
#        """
#        settings.set_symbols_handler(file_extension, symbols_handler)


class MiscContainerService(QObject):

    def __init__(self, miscContainer):
        QObject.__init__(self)
        self._misc = miscContainer

    def add_widget(self, widget, icon_path, description):
        self._misc.add_to_stack(widget, icon_path, description)


class ExplorerService(QObject):
    # SIGNALS
    projectOpened = pyqtSignal("QString")
    projectExecuted = pyqtSignal("QString")

    def __init__(self):
        QObject.__init__(self)
        self._explorer = explorer_container.ExplorerContainer()
        self._action = actions.Actions()
        self.connect(self._explorer, SIGNAL("projectOpened(QString)"),
                     self._projectOpened)
        self.connect(self._action, SIGNAL("projectExecuted(QString)"),
                     self._projectExecuted)

    def get_tree_projects(self):
        """
        Returns the projects tree
        """
        return self._explorer._treeProjects

    def get_item(self, path):
        if self._explorer._treeProjects:
            return self._explorer._treeProjects.get_item_for_path(path)

    def get_current_project_item(self):
        """
        Returns the current item of the tree projects
        this method is a shortcut of self.get_tree_projects().currentItem()
        """
        if self._explorer._treeProjects:
            return self._explorer._treeProjects.currentItem()
        return None

    def get_project_item_by_name(self, projectName):
        """
        Return a ProjectItem that has the name provided.
        """
        if self._explorer._treeProjects:
            return self._explorer._treeProjects.get_project_by_name(
                projectName)
        return None

    def get_tree_symbols(self):
        """
        Returns the symbols tree
        """
        return self._explorer._treeSymbols

    def set_symbols_handler(self, file_extension, symbols_handler):
        """
        Add a new Symbol's handler for the given file extension
        example:
            cpp_symbols_handler = CppSymbolHandler(...)
            set_symbols_handler('cpp', cpp_symbols_handler)
        Then all symbols in .cpp files will be handle by cpp_symbols_handler

        Note: symbols_handler SHOULD have a special interface see
        ninja_ide.core.plugin_interfaces
        """
        settings.set_symbols_handler(file_extension, symbols_handler)

    def set_project_type_handler(self, project_type, project_type_handler):
        """
        Add a new Project Type and the handler for it
        example:
            foo_project_handler = FooProjectHandler(...)
            set_project_type_handler('Foo Project', foo_project_handler)
        Then 'Foo Project' will appear in the New Project wizard
            and foo_project_handler instance controls the wizard

        Note: project_type_handler SHOULD have a special interface see
        ninja_ide.core.plugin_interfaces
        """
        settings.set_project_type_handler(project_type, project_type_handler)

    def add_tab(self, tab, title):
        """
        Add a tab with the given title
        @tab: Should be an instance (or subclass )of QTabWidget
        @title: Name of the tab string
        """
        self._explorer.addTab(tab, title)

    def get_actual_project(self):
        """
        Returns the path of the opened projects
        """
        return self._explorer.get_actual_project()

    def get_opened_projects(self):
        """
        Return the opened projects in the Tree Project Explorer.
        list of <ninja_ide.gui.explorer.tree_projects_widget.ProjectTree>
        """
        opened_projects = self._explorer.get_opened_projects()
        return opened_projects

    def add_project_menu(self, menu, lang='all'):
        """
        Add an extra menu to the project explorer for the files
        with the given extension.
        @lang: String with file extension format (py, php, json)
        """
        if self._explorer._treeProjects:
            self._explorer._treeProjects.add_extra_menu(menu, lang=lang)

    def add_project_menu_by_scope(self, menu, scope=None):
        """
        Add an extra menu to the project explorer to the specific scope
        @scope: String with the menu scope (all, project, folder, file)
        """
        if scope is None:
            #default behavior show ALL
            scope = plugin_util.ContextMenuScope(project=True, folder=True,
                                                 files=True)
        if self._explorer._treeProjects:
            self._explorer._treeProjects.add_extra_menu_by_scope(menu, scope)

    # SIGNALS
    def _projectOpened(self, projectPath):
        """
        Signal emitted when the project is opened
        """
        self.projectOpened.emit(projectPath)

    def _projectExecuted(self, projectPath):
        """
        Signal emitted when the project is executed
        """
        self.projectExecuted.emit(projectPath)



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import webbrowser

from PyQt4.QtGui import QDialog
from PyQt4.QtGui import QVBoxLayout
from PyQt4.QtGui import QHBoxLayout
from PyQt4.QtGui import QLabel
from PyQt4.QtGui import QPixmap
from PyQt4.QtCore import Qt
from PyQt4.QtCore import QSize
from PyQt4.QtCore import SIGNAL

import ninja_ide
from ninja_ide import resources


class AboutNinja(QDialog):

    def __init__(self, parent=None):
        QDialog.__init__(self, parent, Qt.Dialog)
        self.setWindowTitle(self.tr("About NINJA-IDE"))
        self.setMaximumSize(QSize(0, 0))

        vbox = QVBoxLayout(self)

        #Create an icon for the Dialog
        pixmap = QPixmap(resources.IMAGES['icon'])
        self.lblIcon = QLabel()
        self.lblIcon.setPixmap(pixmap)

        hbox = QHBoxLayout()
        hbox.addWidget(self.lblIcon)

        lblTitle = QLabel(
                '<h1>NINJA-IDE</h1>\n<i>Ninja-IDE Is Not Just Another IDE<i>')
        lblTitle.setTextFormat(Qt.RichText)
        lblTitle.setAlignment(Qt.AlignLeft)
        hbox.addWidget(lblTitle)
        vbox.addLayout(hbox)
        #Add description
        vbox.addWidget(QLabel(
self.tr("""NINJA-IDE (from: "Ninja Is Not Just Another IDE"), is a
cross-platform integrated development environment specially designed
to build Python Applications.
NINJA-IDE provides tools to simplify the Python-software development
and handles all kinds of situations thanks to its rich extensibility.""")))
        vbox.addWidget(QLabel(self.tr("Version: %s") % ninja_ide.__version__))
        link_ninja = QLabel(
            self.tr('Website: <a href="%s"><span style=" '
                'text-decoration: underline; color:#ff9e21;">'
                '%s</span></a>') %
                (ninja_ide.__url__, ninja_ide.__url__))
        vbox.addWidget(link_ninja)
        link_source = QLabel(
            self.tr('Source Code: <a href="%s"><span style=" '
            'text-decoration: underline; color:#ff9e21;">%s</span></a>') %
                (ninja_ide.__source__, ninja_ide.__source__))
        vbox.addWidget(link_source)

        self.connect(link_ninja, SIGNAL("linkActivated(QString)"),
            self.link_activated)
        self.connect(link_source, SIGNAL("linkActivated(QString)"),
            self.link_activated)

    def link_activated(self, link):
        webbrowser.open(str(link))



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt4.QtGui import QTabWidget
from PyQt4.QtGui import QCursor
from PyQt4.QtGui import QIcon
from PyQt4.QtGui import QHBoxLayout
from PyQt4.QtGui import QWidget
from PyQt4.QtGui import QMessageBox
from PyQt4.QtGui import QMenu
from PyQt4.QtGui import QPushButton
from PyQt4.QtGui import QApplication
from PyQt4.QtGui import QClipboard
from PyQt4.QtCore import Qt
from PyQt4.QtCore import SIGNAL
from PyQt4.QtCore import QDir

from ninja_ide import resources
from ninja_ide.core import settings
from ninja_ide.core import file_manager
from ninja_ide.core.filesystem_notifications.base_watcher import MODIFIED, \
                                                                DELETED
from ninja_ide.core.filesystem_notifications import NinjaFileSystemWatcher
from ninja_ide.gui.editor import editor
from ninja_ide.gui.main_panel import browser_widget

from ninja_ide.tools.logger import NinjaLogger

logger = NinjaLogger('ninja_ide.gui.main_panel.tab_widget')
DEBUG = logger.debug


class TabWidget(QTabWidget):

###############################################################################
# TabWidget SIGNALS
###############################################################################
    """
    tabCloseRequested(int)
    dropTab(QTabWidget)
    saveActualEditor()
    allTabsClosed()
    changeActualTab(QTabWidget)
    splitTab(QTabWidget, int, bool)
    reopenTab(QTabWidget, QString)
    runFile()
    addToProject(QString)
    syntaxChanged(QWidget, QString)
    reloadFile(QWidget)
    navigateCode(bool, int)
    recentTabsModified(QStringList)
    """
###############################################################################

    def __init__(self, parent):
        QTabWidget.__init__(self, parent)
        self.setTabsClosable(True)
        self.setMovable(True)
        self.setAcceptDrops(True)
        self.notOpening = True
        self.__lastOpened = []
        self._resyntax = []
        self.navigator = TabNavigator()
        self.setCornerWidget(self.navigator, Qt.TopRightCorner)
        self._parent = parent
        self.follow_mode = False
        self._change_map = {}
        #On some platforms there are problem with focusInEvent
        self.question_already_open = False
        #Keep track of the tab titles
        self.titles = []
        self.dontLoopInExpandTitle = False
        self.connect(NinjaFileSystemWatcher,
            SIGNAL("fileChanged(int, QString)"), self._file_changed)
        self.connect(self, SIGNAL("tabCloseRequested(int)"), self.removeTab)
        self.connect(self.navigator.btnPrevious, SIGNAL("clicked()"),
            lambda: self._navigate_code(False))
        self.connect(self.navigator.btnNext, SIGNAL("clicked()"),
            lambda: self._navigate_code(True))

    def get_recent_files_list(self):
        return self.__lastOpened

    def _navigate_code(self, val):
        op = self.navigator.operation
        self.emit(SIGNAL("navigateCode(bool, int)"), val, op)

    def _add_to_last_opened(self, path):
        if path not in self.__lastOpened:
            self.__lastOpened.append(path)
            if len(self.__lastOpened) > settings.MAX_REMEMBER_TABS:
                self.__lastOpened = self.__lastOpened[1:]
            self.emit(SIGNAL("recentTabsModified(QStringList)"),
                self.__lastOpened)

    def add_tab(self, widget, title, index=None):
        try:
            if index is not None:
                inserted_index = self.insertTab(index, widget, title)
            else:
                inserted_index = self.addTab(widget, title)
            self.setCurrentIndex(inserted_index)
            self.expand_tab_name(title)
            widget.setFocus()
            return inserted_index
        except AttributeError as reason:
            msg = "Widget couldn't be added, doesn't inherit from ITabWidget"
            logger.error('add_tab: %s', reason)
            logger.error(msg)

    def expand_tab_name(self, title):
        """Expand the tab title to differentiate files with the same name.

        The way it is currently implemented, it will only change the first
        conflicting title passed in, because it only searches until the new
        title isn't in the tab titles.
        """
        if title == 'New Document':
            return
        elif title not in self.titles:
            self.titles.append(title)
            return
        indexes = [i for i in range(self.count())
            if type(self.widget(i)) is editor.Editor and
            self.tabText(i) == title and
            self.widget(i).ID]  # self.widget.ID returns the basename
        self.dontLoopInExpandTitle = True
        for i in indexes:
            newName = file_manager.create_path(
                file_manager.get_basename(
                    file_manager.get_folder(self.widget(i).ID)), title)
            while newName in self.titles:
                # Keep prepending the folder name onto the title until it
                # does not conflict.
                path = self.widget(i).ID
                tempDir = path[:path.rfind(newName)]
                newName = file_manager.create_path(
                    file_manager.get_basename(
                        file_manager.get_folder(tempDir)),
                    '..',
                    title)
            self.titles.append(newName)
            self.setTabText(i, newName)
        self.dontLoopInExpandTitle = False

    def tab_was_modified(self, val):
        ed = self.currentWidget()
        text = self.tabBar().tabText(self.currentIndex())
        if type(ed) is editor.Editor and self.notOpening and val and \
           not text.startswith('(*) '):
            ed.textModified = True
            text = '(*) %s' % self.tabBar().tabText(self.currentIndex())
            self.tabBar().setTabText(self.currentIndex(), text)

    def focusInEvent(self, event):
        QTabWidget.focusInEvent(self, event)
        self.emit(SIGNAL("changeActualTab(QTabWidget)"), self)

        #custom behavior after the default
        editorWidget = self.currentWidget()
        if not editorWidget:
            return
        #Check never saved
        if editorWidget.newDocument:
            return
        #Check external modifications!
        self.check_for_external_modifications(editorWidget)
        #we can ask again
        self.question_already_open = False

    def _prompt_reload(self, editorWidget, change):
        self.question_already_open = True
        if change == MODIFIED:
            if editorWidget.just_saved:
                editorWidget.just_saved = False
                return
            val = QMessageBox.question(self, 'The file has changed on disk!',
                (self.tr("%s\nDo you want to reload it?") % editorWidget.ID),
                QMessageBox.Yes, QMessageBox.No)
            if val == QMessageBox.Yes:
                self.emit(SIGNAL("reloadFile(QWidget)"), editorWidget)
            else:
                #dont ask again while the current file is open
                editorWidget.ask_if_externally_modified = False
        elif change == DELETED:
                val = QMessageBox.information(self,
                            'The file has been deleted from disk!',
                (self.tr("%s\n") % editorWidget.ID),
                QMessageBox.Yes)
        self.question_already_open = False

    def _file_changed(self, change_type, file_path):
        file_path = QDir.toNativeSeparators(file_path)
        editorWidget = self.currentWidget()
        current_open = QDir.toNativeSeparators(editorWidget and
                                                    editorWidget.ID or "")
        opened = [path for path, _ in self.get_documents_data()]

        if (file_path in opened) and \
            ((not editorWidget) or (current_open != file_path)) and \
            (change_type in (MODIFIED, DELETED)):

            self._change_map.setdefault(file_path,
                                        []).append(change_type)
        elif not editorWidget:
            return
        elif (current_open == file_path) and \
            (not self.question_already_open):
            #dont ask again if you are already asking!
            self._prompt_reload(editorWidget, change_type)

    def check_for_external_modifications(self, editorWidget):
        e_path = editorWidget.ID
        if e_path in self._change_map:
            if DELETED in self._change_map[e_path]:
                self._prompt_reload(editorWidget, DELETED)
            else:
                self._prompt_reload(editorWidget, MODIFIED)
            self._change_map.pop(e_path)

    def tab_was_saved(self, ed):
        index = self.indexOf(ed)
        text = self.tabBar().tabText(index)
        if text.startswith('(*) '):
            text = text[4:]
        self.tabBar().setTabText(index, text)

    def is_open(self, identifier):
        """Check if a Tab with id = identifier is open"""
        for i in range(self.count()):
            if self.widget(i) == identifier:
                return i
        return -1

    def move_to_open(self, identifier):
        """Set the selected Tab for the widget with id = identifier"""
        for i in range(self.count()):
            if self.widget(i) == identifier:
                self.setCurrentIndex(i)
                return

    def search_for_identifier_index(self, identifier):
        for i in range(self.count()):
            if self.widget(i) == identifier:
                return i

    def remove_title(self, index):
        """Looks for the title of the tab at index and removes it from
        self.titles, if it's there.'"""
        if self.tabText(index) in self.titles:
            self.titles.remove(self.tabText(index))

    def update_current_widget(self):
        """Sets the focus to the current widget. If this is the last tab in the
        current split, the allTabsClosed() signal is emitted.'"""
        if self.currentWidget() is not None:
            self.currentWidget().setFocus()
        else:
            self.emit(SIGNAL("allTabsClosed()"))

    def removeTab(self, index):
        """Remove the Tab at the selected index and check if the
        widget was modified and need to execute any saving"""
        if index != -1:
            self.setCurrentIndex(index)
            widget = self.currentWidget()
            if type(widget) is editor.Editor:
                val = QMessageBox.No
                if widget.textModified and not self.follow_mode:
                    fileName = self.tabBar().tabText(self.currentIndex())
                    val = QMessageBox.question(
                        self, (self.tr('The file %s was not saved') %
                            fileName),
                            self.tr("Do you want to save before closing?"),
                            QMessageBox.Yes | QMessageBox.No |
                            QMessageBox.Cancel)
                if val == QMessageBox.Cancel:
                    return
                elif val == QMessageBox.Yes:
                    self.emit(SIGNAL("saveActualEditor()"))
                    if widget.textModified:
                        return
            if type(widget) == browser_widget.BrowserWidget:
                widget.shutdown_pydoc()
            elif type(widget) is editor.Editor and widget.ID:
                self._add_to_last_opened(widget.ID)
                self._parent.remove_standalone_watcher(widget.ID)
                widget.completer.cc.unload_module()

            self.remove_title(index)
            super(TabWidget, self).removeTab(index)
            del widget
            self.update_current_widget()

    def setTabText(self, index, text):
        QTabWidget.setTabText(self, index, text)
        if text in self.titles and not self.dontLoopInExpandTitle:
            self.expand_tab_name(text)

    def close_tab(self):
        self.removeTab(self.currentIndex())

    def get_documents_data(self):
        """Return Editors: path, project, cursor position"""
        files = []
        for i in range(self.count()):
            if (type(self.widget(i)) is editor.Editor) \
            and self.widget(i).ID != '':
                files.append([self.widget(i).ID,
                    self.widget(i).get_cursor_position()])
                self.widget(i)._sidebarWidget._save_breakpoints_bookmarks()
        return files

    def mousePressEvent(self, event):
        QTabWidget.mousePressEvent(self, event)
        if self.follow_mode:
            return
        if event.button() == Qt.RightButton:
            index = self.tabBar().tabAt(event.pos())
            self.setCurrentIndex(index)
            widget = self.widget(index)
            if type(widget) is editor.Editor:
                #show menu
                menu = QMenu()
                actionAdd = menu.addAction(self.tr("Add to Project..."))
                actionRun = menu.addAction(self.tr("Run this File!"))
                menuSyntax = menu.addMenu(self.tr("Change Syntax"))
                self._create_menu_syntax(menuSyntax)
                menu.addSeparator()
                actionClose = menu.addAction(self.tr("Close This Tab"))
                actionCloseAll = menu.addAction(self.tr("Close All Tabs"))
                actionCloseAllNotThis = menu.addAction(
                    self.tr("Close Other Tabs"))
                menu.addSeparator()
                if self._parent.splitted:
                    actionMoveSplit = menu.addAction(
                        self.tr("Move this Tab to the other Split"))
                    actionCloseSplit = menu.addAction(
                        self.tr("Close Split"))
                    #Connect split actions
                    self.connect(actionMoveSplit, SIGNAL("triggered()"),
                        lambda: self._parent.move_tab_to_next_split(self))
                    self.connect(actionCloseSplit, SIGNAL("triggered()"),
                        lambda: self._parent.split_tab(
                            self._parent.orientation() == Qt.Horizontal))
                else:
                    actionSplitH = menu.addAction(
                        self.tr("Split this Tab (Vertically)"))
                    actionSplitV = menu.addAction(
                        self.tr("Split this Tab (Horizontally)"))
                    #Connect split actions
                    self.connect(actionSplitH, SIGNAL("triggered()"),
                        lambda: self._split_this_tab(True))
                    self.connect(actionSplitV, SIGNAL("triggered()"),
                        lambda: self._split_this_tab(False))
                menu.addSeparator()
                actionCopyPath = menu.addAction(
                    self.tr("Copy file location to Clipboard"))
                actionReopen = menu.addAction(
                    self.tr("Reopen last closed File"))
                if len(self.__lastOpened) == 0:
                    actionReopen.setEnabled(False)
                #Connect actions
                self.connect(actionRun, SIGNAL("triggered()"),
                    self._run_this_file)
                self.connect(actionAdd, SIGNAL("triggered()"),
                    self._add_to_project)
                self.connect(actionClose, SIGNAL("triggered()"),
                    lambda: self.removeTab(index))
                self.connect(actionCloseAllNotThis, SIGNAL("triggered()"),
                    self._close_all_tabs_except_this)
                self.connect(actionCloseAll, SIGNAL("triggered()"),
                    self._close_all_tabs)
                self.connect(actionCopyPath, SIGNAL("triggered()"),
                    self._copy_file_location)
                self.connect(actionReopen, SIGNAL("triggered()"),
                    self._reopen_last_tab)
                menu.exec_(event.globalPos())
        if event.button() == Qt.MidButton:
            index = self.tabBar().tabAt(event.pos())
            self.removeTab(index)

    def _create_menu_syntax(self, menuSyntax):
        syntax = list(settings.SYNTAX.keys())
        syntax.sort()
        for syn in syntax:
            menuSyntax.addAction(syn)
            self.connect(menuSyntax, SIGNAL("triggered(QAction*)"),
                self._reapply_syntax)

    def _reapply_syntax(self, syntaxAction):
        if [self.currentIndex(), syntaxAction] != self._resyntax:
            self._resyntax = [self.currentIndex(), syntaxAction]
            self.emit(SIGNAL("syntaxChanged(QWidget, QString)"),
                self.currentWidget(), syntaxAction.text())

    def _run_this_file(self):
        self.emit(SIGNAL("runFile()"))

    def _add_to_project(self):
        self.emit(SIGNAL("changeActualTab(QTabWidget)"), self)
        widget = self.currentWidget()
        if type(widget) is editor.Editor:
            self.emit(SIGNAL("addToProject(QString)"), widget.ID)

    def _reopen_last_tab(self):
        self.emit(SIGNAL("reopenTab(QTabWidget, QString)"),
            self, self.__lastOpened.pop())
        self.emit(SIGNAL("recentTabsModified(QStringList)"), self.__lastOpened)

    def _split_this_tab(self, orientation):
        self.emit(SIGNAL("splitTab(QTabWidget, int, bool)"),
            self, self.currentIndex(), orientation)

    def _copy_file_location(self):
        widget = self.currentWidget()
        QApplication.clipboard().setText(widget.ID, QClipboard.Clipboard)

    def _close_all_tabs(self):
        for i in range(self.count()):
            self.removeTab(0)

    def _close_all_tabs_except_this(self):
        self.tabBar().moveTab(self.currentIndex(), 0)
        for i in range(self.count()):
            if self.count() > 1:
                self.removeTab(1)

    def _check_unsaved_tabs(self):
        """
        Check if are there any unsaved tab
        Returns True or False
        """
        val = False
        for i in range(self.count()):
            if type(self.widget(i)) is editor.Editor:
                val = val or self.widget(i).textModified
        return val

    def get_unsaved_files(self):
        """
        Returns a list with the tabText of the unsaved files
        """
        files = []
        for i in range(self.count()):
            widget = self.widget(i)
            if type(widget) is editor.Editor and widget.textModified:
                files.append(self.tabText(i))
        return files

    def change_tab(self):
        if self.currentIndex() < (self.count() - 1):
            self.setCurrentIndex(self.currentIndex() + 1)
        else:
            self.setCurrentIndex(0)

    def change_tab_reverse(self):
        if self.currentIndex() > 0:
            self.setCurrentIndex(self.currentIndex() - 1)
        else:
            self.setCurrentIndex(self.count() - 1)

    def change_open_tab_name(self, index, newName):
        """Change the name of the tab at index, for the newName."""
        self.remove_title(index)
        self.setTabText(index, newName)
        self.titles.append(newName)


class TabNavigator(QWidget):

    def __init__(self):
        QWidget.__init__(self)
        self.setContextMenuPolicy(Qt.DefaultContextMenu)
        self.setMinimumHeight(38)
        hbox = QHBoxLayout(self)
        self.btnPrevious = QPushButton(
            QIcon(resources.IMAGES['nav-code-left']), '')
        self.btnPrevious.setObjectName('navigation_button')
        self.btnPrevious.setToolTip(
            self.tr("Right click to change navigation options"))
        self.btnNext = QPushButton(
            QIcon(resources.IMAGES['nav-code-right']), '')
        self.btnNext.setObjectName('navigation_button')
        self.btnNext.setToolTip(
            self.tr("Right click to change navigation options"))
        hbox.addWidget(self.btnPrevious)
        hbox.addWidget(self.btnNext)
        self.setContentsMargins(0, 0, 0, 0)

        self.menuNavigate = QMenu(self.tr("Navigate"))
        self.codeAction = self.menuNavigate.addAction(
            self.tr("Code Jumps"))
        self.codeAction.setCheckable(True)
        self.codeAction.setChecked(True)
        self.bookmarksAction = self.menuNavigate.addAction(
            self.tr("Bookmarks"))
        self.bookmarksAction.setCheckable(True)
        self.breakpointsAction = self.menuNavigate.addAction(
            self.tr("Breakpoints"))
        self.breakpointsAction.setCheckable(True)

        # 0 = Code Jumps
        # 1 = Bookmarks
        # 2 = Breakpoints
        self.operation = 0

        self.connect(self.codeAction, SIGNAL("triggered()"),
            self._show_code_nav)
        self.connect(self.breakpointsAction, SIGNAL("triggered()"),
            self._show_breakpoints)
        self.connect(self.bookmarksAction, SIGNAL("triggered()"),
            self._show_bookmarks)

    def contextMenuEvent(self, event):
        self.show_menu_navigation()

    def show_menu_navigation(self):
        self.menuNavigate.exec_(QCursor.pos())

    def _show_bookmarks(self):
        self.btnPrevious.setIcon(QIcon(resources.IMAGES['book-left']))
        self.btnNext.setIcon(QIcon(resources.IMAGES['book-right']))
        self.bookmarksAction.setChecked(True)
        self.breakpointsAction.setChecked(False)
        self.codeAction.setChecked(False)
        self.operation = 1

    def _show_breakpoints(self):
        self.btnPrevious.setIcon(QIcon(resources.IMAGES['break-left']))
        self.btnNext.setIcon(QIcon(resources.IMAGES['break-right']))
        self.bookmarksAction.setChecked(False)
        self.breakpointsAction.setChecked(True)
        self.codeAction.setChecked(False)
        self.operation = 2

    def _show_code_nav(self):
        self.btnPrevious.setIcon(QIcon(resources.IMAGES['nav-code-left']))
        self.btnNext.setIcon(QIcon(resources.IMAGES['nav-code-right']))
        self.bookmarksAction.setChecked(False)
        self.breakpointsAction.setChecked(False)
        self.codeAction.setChecked(True)
        self.operation = 0



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt4.QtCore import QObject
from PyQt4.QtCore import SIGNAL

from ninja_ide.gui import central_widget
from ninja_ide.gui.dialogs import plugins_manager
from ninja_ide.gui.dialogs import themes_manager
from ninja_ide.gui.dialogs import language_manager


class MenuPlugins(QObject):

    def __init__(self, menuPlugins):
        QObject.__init__(self)

        manageAction = menuPlugins.addAction(self.tr("Manage Plugins"))
        skinsAction = menuPlugins.addAction(self.tr("Editor Schemes"))
        languagesAction = menuPlugins.addAction(self.tr("Languages Manager"))
        menuPlugins.addSeparator()

        self.connect(manageAction, SIGNAL("triggered()"), self._show_manager)
        self.connect(skinsAction, SIGNAL("triggered()"), self._show_themes)
        self.connect(languagesAction, SIGNAL("triggered()"),
            self._show_languages)

    def _show_manager(self):
        manager = plugins_manager.PluginsManagerWidget(
            central_widget.CentralWidget())
        manager.exec_()
        if manager._requirements:
            d = plugins_manager.DependenciesHelpDialog(manager._requirements)
            d.exec_()

    def _show_languages(self):
        manager = language_manager.LanguagesManagerWidget(
            central_widget.CentralWidget())
        manager.show()

    def _show_themes(self):
        manager = themes_manager.ThemesManagerWidget(
            central_widget.CentralWidget())
        manager.show()



#!/usr/bin/env python
# pep8.py - Check Python source code formatting, according to PEP 8
# Copyright (C) 2006 Johann C. Rocholl <johann@rocholl.net>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

r"""
Check Python source code formatting, according to PEP 8:
http://www.python.org/dev/peps/pep-0008/

For usage and a list of options, try this:
$ python pep8.py -h

This program and its regression test suite live here:
http://github.com/jcrocholl/pep8

Groups of errors and warnings:
E errors
W warnings
100 indentation
200 whitespace
300 blank lines
400 imports
500 line length
600 deprecation
700 statements
900 syntax error

You can add checks to this program by writing plugins. Each plugin is
a simple function that is called for each line of source code, either
physical or logical.

Physical line:
- Raw line of text from the input file.

Logical line:
- Multi-line statements converted to a single line.
- Stripped left and right.
- Contents of strings replaced with 'xxx' of same length.
- Comments removed.

The check function requests physical or logical lines by the name of
the first argument:

def maximum_line_length(physical_line)
def extraneous_whitespace(logical_line)
def blank_lines(logical_line, blank_lines, indent_level, line_number)

The last example above demonstrates how check plugins can request
additional information with extra arguments. All attributes of the
Checker object are available. Some examples:

lines: a list of the raw lines from the input file
tokens: the tokens that contribute to this logical line
line_number: line number in the input file
blank_lines: blank lines before this one
indent_char: first indentation character in this file (' ' or '\t')
indent_level: indentation (with tabs expanded to multiples of 8)
previous_indent_level: indentation on previous line
previous_logical: previous logical line

The docstring of each check function shall be the relevant part of
text from PEP 8. It is printed if the user enables --show-pep8.
Several docstrings contain examples directly from the PEP 8 document.

Okay: spam(ham[1], {eggs: 2})
E201: spam( ham[1], {eggs: 2})

These examples are verified automatically when pep8.py is run with the
--doctest option. You can add examples for your own check functions.
The format is simple: "Okay" or error/warning code followed by colon
and space, the rest of the line is example source code. If you put 'r'
before the docstring, you can use \n for newline, \t for tab and \s
for space.

"""
from __future__ import unicode_literals

__version__ = '1.3.4a0'

import os
import sys
import re
import inspect
import keyword
import tokenize

from ninja_ide import resources

DEFAULT_EXCLUDE = '.svn,CVS,.bzr,.hg,.git'
DEFAULT_IGNORE = 'E24'
if sys.platform == 'win32':
    DEFAULT_CONFIG = os.path.join(resources.HOME_PATH, r'.pep8')
else:
    DEFAULT_CONFIG = os.path.join(os.getenv('XDG_CONFIG_HOME') or
                                  os.path.join(
                                      resources.HOME_PATH, '.config'), 'pep8')
MAX_LINE_LENGTH = 79
REPORT_FORMAT = {
    'default': '%(path)s:%(row)d:%(col)d: %(code)s %(text)s',
    'pylint': '%(path)s:%(row)d: [%(code)s] %(text)s',
}


SINGLETONS = frozenset(['False', 'None', 'True'])
KEYWORDS = frozenset(keyword.kwlist + ['print']) - SINGLETONS
BINARY_OPERATORS = frozenset([
    '**=', '*=', '+=', '-=', '!=', '<>',
    '%=', '^=', '&=', '|=', '==', '/=', '//=', '<=', '>=', '<<=', '>>=',
    '%', '^', '&', '|', '=', '/', '//', '<', '>', '<<'])
UNARY_OPERATORS = frozenset(['>>', '**', '*', '+', '-'])
OPERATORS = BINARY_OPERATORS | UNARY_OPERATORS
WHITESPACE = frozenset(' \t')
SKIP_TOKENS = frozenset([tokenize.COMMENT, tokenize.NL, tokenize.NEWLINE,
                         tokenize.INDENT, tokenize.DEDENT])
BENCHMARK_KEYS = ['directories', 'files', 'logical lines', 'physical lines']

INDENT_REGEX = re.compile(r'([ \t]*)')
RAISE_COMMA_REGEX = re.compile(r'raise\s+\w+\s*(,)')
RERAISE_COMMA_REGEX = re.compile(r'raise\s+\w+\s*,\s*\w+\s*,\s*\w+')
SELFTEST_REGEX = re.compile(r'(Okay|[EW]\d{3}):\s(.*)')
ERRORCODE_REGEX = re.compile(r'[EW]\d{3}')
DOCSTRING_REGEX = re.compile(r'u?r?["\']')
EXTRANEOUS_WHITESPACE_REGEX = re.compile(r'[[({] | []}),;:]')
WHITESPACE_AFTER_COMMA_REGEX = re.compile(r'[,;:]\s*(?:  |\t)')
COMPARE_SINGLETON_REGEX = re.compile(r'([=!]=)\s*(None|False|True)')
COMPARE_TYPE_REGEX = re.compile(r'([=!]=|is|is\s+not)\s*type(?:s\.(\w+)Type'
                                r'|\(\s*(\(\s*\)|[^)]*[^ )])\s*\))')
KEYWORD_REGEX = re.compile(r'(?:[^\s])(\s*)\b(?:%s)\b(\s*)' %
                           r'|'.join(KEYWORDS))
OPERATOR_REGEX = re.compile(r'(?:[^\s])(\s*)(?:[-+*/|!<=>%&^]+)(\s*)')
LAMBDA_REGEX = re.compile(r'\blambda\b')
HUNK_REGEX = re.compile(r'^@@ -\d+,\d+ \+(\d+),(\d+) @@.*$')

# Work around Python < 2.6 behaviour, which does not generate NL after
# a comment which is on a line by itself.
COMMENT_WITH_NL = tokenize.generate_tokens(['#\n'].pop).send(None)[1] == '#\n'


#Define instance options to set options to be analized
class Options:
    pass


options = Options()
#Set the options to be analized
options.show_source = True
options.repeat = True
options.show_pep8 = False
options.messages = {}
options.select = []
options.ignore = []
options.verbose = 0
options.quiet = 0
options.max_line_length = MAX_LINE_LENGTH
options.ignore_continuation_indentation = True


##############################################################################
# Plugins (check functions) for physical lines
##############################################################################


def tabs_or_spaces(physical_line, indent_char):
    r"""
    Never mix tabs and spaces.

    The most popular way of indenting Python is with spaces only.  The
    second-most popular way is with tabs only.  Code indented with a mixture
    of tabs and spaces should be converted to using spaces exclusively.  When
    invoking the Python command line interpreter with the -t option, it issues
    warnings about code that illegally mixes tabs and spaces.  When using -tt
    these warnings become errors.  These options are highly recommended!

    Okay: if a == 0:\n        a = 1\n        b = 1
    E101: if a == 0:\n        a = 1\n\tb = 1
    """
    indent = INDENT_REGEX.match(physical_line).group(1)
    for offset, char in enumerate(indent):
        if char != indent_char:
            return offset, "E101 indentation contains mixed spaces and tabs"


def tabs_obsolete(physical_line):
    r"""
    For new projects, spaces-only are strongly recommended over tabs.  Most
    editors have features that make this easy to do.

    Okay: if True:\n    return
    W191: if True:\n\treturn
    """
    indent = INDENT_REGEX.match(physical_line).group(1)
    if '\t' in indent:
        return indent.index('\t'), "W191 indentation contains tabs"


def trailing_whitespace(physical_line):
    r"""
    JCR: Trailing whitespace is superfluous.
    FBM: Except when it occurs as part of a blank line (i.e. the line is
         nothing but whitespace). According to Python docs[1] a line with only
         whitespace is considered a blank line, and is to be ignored. However,
         matching a blank line to its indentation level avoids mistakenly
         terminating a multi-line statement (e.g. class declaration) when
         pasting code into the standard Python interpreter.

         [1] http://docs.python.org/reference/lexical_analysis.html#blank-lines

    The warning returned varies on whether the line itself is blank, for easier
    filtering for those who want to indent their blank lines.

    Okay: spam(1)
    W291: spam(1)\s
    W293: class Foo(object):\n    \n    bang = 12
    """
    physical_line = physical_line.rstrip('\n')    # chr(10), newline
    physical_line = physical_line.rstrip('\r')    # chr(13), carriage return
    physical_line = physical_line.rstrip('\x0c')  # chr(12), form feed, ^L
    stripped = physical_line.rstrip(' \t\v')
    if physical_line != stripped:
        if stripped:
            return len(stripped), "W291 trailing whitespace"
        else:
            return 0, "W293 blank line contains whitespace"


def trailing_blank_lines(physical_line, lines, line_number):
    r"""
    JCR: Trailing blank lines are superfluous.

    Okay: spam(1)
    W391: spam(1)\n
    """
    if not physical_line.rstrip() and line_number == len(lines):
        return 0, "W391 blank line at end of file"


def missing_newline(physical_line):
    """
    JCR: The last line should have a newline.

    Reports warning W292.
    """
    if physical_line.rstrip() == physical_line:
        return len(physical_line), "W292 no newline at end of file"


def maximum_line_length(physical_line, max_line_length):
    """
    Limit all lines to a maximum of 79 characters.

    There are still many devices around that are limited to 80 character
    lines; plus, limiting windows to 80 characters makes it possible to have
    several windows side-by-side.  The default wrapping on such devices looks
    ugly.  Therefore, please limit all lines to a maximum of 79 characters.
    For flowing long blocks of text (docstrings or comments), limiting the
    length to 72 characters is recommended.

    Reports error E501.
    """
    line = physical_line.rstrip()
    length = len(line)
    if length > max_line_length:
        if hasattr(line, 'decode'):   # Python 2
            # The line could contain multi-byte characters
            try:
                length = len(line.decode('utf-8'))
            except UnicodeError:
                pass
        if length > max_line_length:
            return (max_line_length, "E501 line too long "
                    "(%d > %d characters)" % (length, max_line_length))


##############################################################################
# Plugins (check functions) for logical lines
##############################################################################


def blank_lines(logical_line, blank_lines, indent_level, line_number,
                previous_logical, previous_indent_level):
    r"""
    Separate top-level function and class definitions with two blank lines.

    Method definitions inside a class are separated by a single blank line.

    Extra blank lines may be used (sparingly) to separate groups of related
    functions.  Blank lines may be omitted between a bunch of related
    one-liners (e.g. a set of dummy implementations).

    Use blank lines in functions, sparingly, to indicate logical sections.

    Okay: def a():\n    pass\n\n\ndef b():\n    pass
    Okay: def a():\n    pass\n\n\n# Foo\n# Bar\n\ndef b():\n    pass

    E301: class Foo:\n    b = 0\n    def bar():\n        pass
    E302: def a():\n    pass\n\ndef b(n):\n    pass
    E303: def a():\n    pass\n\n\n\ndef b(n):\n    pass
    E303: def a():\n\n\n\n    pass
    E304: @decorator\n\ndef a():\n    pass
    """
    if line_number == 1:
        return  # Don't expect blank lines before the first line
    if previous_logical.startswith('@'):
        if blank_lines:
            yield 0, "E304 blank lines found after function decorator"
    elif blank_lines > 2 or (indent_level and blank_lines == 2):
        yield 0, "E303 too many blank lines (%d)" % blank_lines
    elif logical_line.startswith(('def ', 'class ', '@')):
        if indent_level:
            if not (blank_lines or previous_indent_level < indent_level or
                    DOCSTRING_REGEX.match(previous_logical)):
                yield 0, "E301 expected 1 blank line, found 0"
        elif blank_lines != 2:
            yield 0, "E302 expected 2 blank lines, found %d" % blank_lines


def extraneous_whitespace(logical_line):
    """
    Avoid extraneous whitespace in the following situations:

    - Immediately inside parentheses, brackets or braces.

    - Immediately before a comma, semicolon, or colon.

    Okay: spam(ham[1], {eggs: 2})
    E201: spam( ham[1], {eggs: 2})
    E201: spam(ham[ 1], {eggs: 2})
    E201: spam(ham[1], { eggs: 2})
    E202: spam(ham[1], {eggs: 2} )
    E202: spam(ham[1 ], {eggs: 2})
    E202: spam(ham[1], {eggs: 2 })

    E203: if x == 4: print x, y; x, y = y , x
    E203: if x == 4: print x, y ; x, y = y, x
    E203: if x == 4 : print x, y; x, y = y, x
    """
    line = logical_line
    for match in EXTRANEOUS_WHITESPACE_REGEX.finditer(line):
        text = match.group()
        char = text.strip()
        found = match.start()
        if text == char + ' ':
            # assert char in '([{'
            yield found + 1, "E201 whitespace after '%s'" % char
        elif line[found - 1] != ',':
            code = ('E202' if char in '}])' else 'E203')  # if char in ',;:'
            yield found, "%s whitespace before '%s'" % (code, char)


def whitespace_around_keywords(logical_line):
    r"""
    Avoid extraneous whitespace around keywords.

    Okay: True and False
    E271: True and  False
    E272: True  and False
    E273: True and\tFalse
    E274: True\tand False
    """
    for match in KEYWORD_REGEX.finditer(logical_line):
        before, after = match.groups()

        if '\t' in before:
            yield match.start(1), "E274 tab before keyword"
        elif len(before) > 1:
            yield match.start(1), "E272 multiple spaces before keyword"

        if '\t' in after:
            yield match.start(2), "E273 tab after keyword"
        elif len(after) > 1:
            yield match.start(2), "E271 multiple spaces after keyword"


def missing_whitespace(logical_line):
    """
    JCR: Each comma, semicolon or colon should be followed by whitespace.

    Okay: [a, b]
    Okay: (3,)
    Okay: a[1:4]
    Okay: a[:4]
    Okay: a[1:]
    Okay: a[1:4:2]
    E231: ['a','b']
    E231: foo(bar,baz)
    """
    line = logical_line
    for index in range(len(line) - 1):
        char = line[index]
        if char in ',;:' and line[index + 1] not in WHITESPACE:
            before = line[:index]
            if char == ':' and before.count('[') > before.count(']'):
                continue  # Slice syntax, no space required
            if char == ',' and line[index + 1] == ')':
                continue  # Allow tuple with only one element: (3,)
            yield index, "E231 missing whitespace after '%s'" % char


def indentation(logical_line, previous_logical, indent_char,
                indent_level, previous_indent_level):
    r"""
    Use 4 spaces per indentation level.

    For really old code that you don't want to mess up, you can continue to
    use 8-space tabs.

    Okay: a = 1
    Okay: if a == 0:\n    a = 1
    E111:   a = 1

    Okay: for item in items:\n    pass
    E112: for item in items:\npass

    Okay: a = 1\nb = 2
    E113: a = 1\n    b = 2
    """
    if indent_char == ' ' and indent_level % 4:
        yield 0, "E111 indentation is not a multiple of four"
    indent_expect = previous_logical.endswith(':')
    if indent_expect and indent_level <= previous_indent_level:
        yield 0, "E112 expected an indented block"
    if indent_level > previous_indent_level and not indent_expect:
        yield 0, "E113 unexpected indentation"


def continuation_line_indentation(logical_line, tokens, indent_level, verbose):
    r"""
    Continuation lines should align wrapped elements either vertically using
    Python's implicit line joining inside parentheses, brackets and braces, or
    using a hanging indent.

    When using a hanging indent the following considerations should be applied:

    - there should be no arguments on the first line, and

    - further indentation should be used to clearly distinguish itself as a
      continuation line.

    Okay: a = (\n)
    E123: a = (\n    )

    Okay: a = (\n    42)
    E121: a = (\n   42)
    E122: a = (\n42)
    E123: a = (\n    42\n    )
    E124: a = (24,\n     42\n)
    E125: if (a or\n    b):\n    pass
    E126: a = (\n        42)
    E127: a = (24,\n      42)
    E128: a = (24,\n    42)
    """
    if options.ignore_continuation_indentation:
        return
    first_row = tokens[0][2][0]
    nrows = 1 + tokens[-1][2][0] - first_row
    if nrows == 1:
        return

    # indent_next tells us whether the next block is indented; assuming
    # that it is indented by 4 spaces, then we should not allow 4-space
    # indents on the final continuation line; in turn, some other
    # indents are allowed to have an extra 4 spaces.
    indent_next = logical_line.endswith(':')

    row = depth = 0
    # remember how many brackets were opened on each line
    parens = [0] * nrows
    # relative indents of physical lines
    rel_indent = [0] * nrows
    # visual indents
    indent = [indent_level]
    indent_chances = {}
    last_indent = tokens[0][2]

    for token_type, text, start, end, line in tokens:
        newline = row < start[0] - first_row
        if newline:
            row = start[0] - first_row
            newline = (not last_token_multiline and
                       token_type not in (tokenize.NL, tokenize.NEWLINE))

        if newline:
            # this is the beginning of a continuation line.
            last_indent = start

            # record the initial indent.
            rel_indent[row] = start[1] - indent_level

            if depth:
                # a bracket expression in a continuation line.
                # find the line that it was opened on
                for open_row in range(row - 1, -1, -1):
                    if parens[open_row]:
                        break
            else:
                # an unbracketed continuation line (ie, backslash)
                open_row = 0
            hang = rel_indent[row] - rel_indent[open_row]
            visual_indent = indent_chances.get(start[1])

            if token_type == tokenize.OP and text in ']})':
                # this line starts with a closing bracket
                if indent[depth]:
                    if start[1] != indent[depth]:
                        yield (start, 'E124 closing bracket does not match '
                               'visual indentation')
                elif hang:
                    yield (start, 'E123 closing bracket does not match '
                           'indentation of opening bracket\'s line')
            elif visual_indent is True:
                # visual indent is verified
                if not indent[depth]:
                    indent[depth] = start[1]
            elif visual_indent in (text, str):
                # ignore token lined up with matching one from a previous line
                pass
            elif indent[depth] and start[1] < indent[depth]:
                # visual indent is broken
                yield (start, 'E128 continuation line '
                       'under-indented for visual indent')
            elif hang == 4 or (indent_next and rel_indent[row] == 8):
                # hanging indent is verified
                pass
            else:
                # indent is broken
                if hang <= 0:
                    error = 'E122', 'missing indentation or outdented'
                elif indent[depth]:
                    error = 'E127', 'over-indented for visual indent'
                elif hang % 4:
                    error = 'E121', 'indentation is not a multiple of four'
                else:
                    error = 'E126', 'over-indented for hanging indent'
                yield start, "%s continuation line %s" % error

        # look for visual indenting
        if (parens[row] and token_type not in (tokenize.NL, tokenize.COMMENT)
                and not indent[depth]):
            indent[depth] = start[1]
            indent_chances[start[1]] = True
        # deal with implicit string concatenation
        elif token_type == tokenize.STRING or text in ('u', 'ur', 'b', 'br'):
            indent_chances[start[1]] = str

        # keep track of bracket depth
        if token_type == tokenize.OP:
            if text in '([{':
                depth += 1
                indent.append(0)
                parens[row] += 1
            elif text in ')]}' and depth > 0:
                # parent indents should not be more than this one
                prev_indent = indent.pop() or last_indent[1]
                for d in range(depth):
                    if indent[d] > prev_indent:
                        indent[d] = 0
                for ind in list(indent_chances):
                    if ind >= prev_indent:
                        del indent_chances[ind]
                depth -= 1
                if depth:
                    indent_chances[indent[depth]] = True
                for idx in range(row, -1, -1):
                    if parens[idx]:
                        parens[idx] -= 1
                        break
            assert len(indent) == depth + 1
            if start[1] not in indent_chances:
                # allow to line up tokens
                indent_chances[start[1]] = text

        last_token_multiline = (start[0] != end[0])

    if indent_next and rel_indent[-1] == 4:
        yield (last_indent, "E125 continuation line does not distinguish "
               "itself from next logical line")


def whitespace_before_parameters(logical_line, tokens):
    """
    Avoid extraneous whitespace in the following situations:

    - Immediately before the open parenthesis that starts the argument
      list of a function call.

    - Immediately before the open parenthesis that starts an indexing or
      slicing.

    Okay: spam(1)
    E211: spam (1)

    Okay: dict['key'] = list[index]
    E211: dict ['key'] = list[index]
    E211: dict['key'] = list [index]
    """
    prev_type = tokens[0][0]
    prev_text = tokens[0][1]
    prev_end = tokens[0][3]
    for index in range(1, len(tokens)):
        token_type, text, start, end, line = tokens[index]
        if (token_type == tokenize.OP and
            text in '([' and
            start != prev_end and
            (prev_type == tokenize.NAME or prev_text in '}])') and
            # Syntax "class A (B):" is allowed, but avoid it
            (index < 2 or tokens[index - 2][1] != 'class') and
                # Allow "return (a.foo for a in range(5))"
                not keyword.iskeyword(prev_text)):
            yield prev_end, "E211 whitespace before '%s'" % text
        prev_type = token_type
        prev_text = text
        prev_end = end


def whitespace_around_operator(logical_line):
    r"""
    Avoid extraneous whitespace in the following situations:

    - More than one space around an assignment (or other) operator to
      align it with another.

    Okay: a = 12 + 3
    E221: a = 4  + 5
    E222: a = 4 +  5
    E223: a = 4\t+ 5
    E224: a = 4 +\t5
    """
    for match in OPERATOR_REGEX.finditer(logical_line):
        before, after = match.groups()

        if '\t' in before:
            yield match.start(1), "E223 tab before operator"
        elif len(before) > 1:
            yield match.start(1), "E221 multiple spaces before operator"

        if '\t' in after:
            yield match.start(2), "E224 tab after operator"
        elif len(after) > 1:
            yield match.start(2), "E222 multiple spaces after operator"


def missing_whitespace_around_operator(logical_line, tokens):
    r"""
    - Always surround these binary operators with a single space on
      either side: assignment (=), augmented assignment (+=, -= etc.),
      comparisons (==, <, >, !=, <>, <=, >=, in, not in, is, is not),
      Booleans (and, or, not).

    - Use spaces around arithmetic operators.

    Okay: i = i + 1
    Okay: submitted += 1
    Okay: x = x * 2 - 1
    Okay: hypot2 = x * x + y * y
    Okay: c = (a + b) * (a - b)
    Okay: foo(bar, key='word', *args, **kwargs)
    Okay: baz(**kwargs)
    Okay: negative = -1
    Okay: spam(-1)
    Okay: alpha[:-i]
    Okay: if not -5 < x < +5:\n    pass
    Okay: lambda *args, **kw: (args, kw)

    E225: i=i+1
    E225: submitted +=1
    E225: x = x*2 - 1
    E225: hypot2 = x*x + y*y
    E225: c = (a+b) * (a-b)
    E225: c = alpha -4
    E225: z = x **y
    """
    parens = 0
    need_space = False
    prev_type = tokenize.OP
    prev_text = prev_end = None
    for token_type, text, start, end, line in tokens:
        if token_type in (tokenize.NL, tokenize.NEWLINE, tokenize.ERRORTOKEN):
            # ERRORTOKEN is triggered by backticks in Python 3000
            continue
        if text in ('(', 'lambda'):
            parens += 1
        elif text == ')':
            parens -= 1
        if need_space:
            if start != prev_end:
                need_space = False
            elif text == '>' and prev_text in ('<', '-'):
                # Tolerate the "<>" operator, even if running Python 3
                # Deal with Python 3's annotated return value "->"
                pass
            else:
                yield prev_end, "E225 missing whitespace around operator"
                need_space = False
        elif token_type == tokenize.OP and prev_end is not None:
            if text == '=' and parens:
                # Allow keyword args or defaults: foo(bar=None).
                pass
            elif text in BINARY_OPERATORS:
                need_space = True
            elif text in UNARY_OPERATORS:
                # Allow unary operators: -123, -x, +1.
                # Allow argument unpacking: foo(*args, **kwargs).
                if prev_type == tokenize.OP:
                    if prev_text in '}])':
                        need_space = True
                elif prev_type == tokenize.NAME:
                    if prev_text not in KEYWORDS:
                        need_space = True
                elif prev_type not in SKIP_TOKENS:
                    need_space = True
            if need_space and start == prev_end:
                yield prev_end, "E225 missing whitespace around operator"
                need_space = False
        prev_type = token_type
        prev_text = text
        prev_end = end


def whitespace_around_comma(logical_line):
    r"""
    Avoid extraneous whitespace in the following situations:

    - More than one space around an assignment (or other) operator to
      align it with another.

    Note: these checks are disabled by default

    Okay: a = (1, 2)
    E241: a = (1,  2)
    E242: a = (1,\t2)
    """
    line = logical_line
    for m in WHITESPACE_AFTER_COMMA_REGEX.finditer(line):
        found = m.start() + 1
        if '\t' in m.group():
            yield found, "E242 tab after '%s'" % m.group()[0]
        else:
            yield found, "E241 multiple spaces after '%s'" % m.group()[0]


def whitespace_around_named_parameter_equals(logical_line, tokens):
    """
    Don't use spaces around the '=' sign when used to indicate a
    keyword argument or a default parameter value.

    Okay: def complex(real, imag=0.0):
    Okay: return magic(r=real, i=imag)
    Okay: boolean(a == b)
    Okay: boolean(a != b)
    Okay: boolean(a <= b)
    Okay: boolean(a >= b)

    E251: def complex(real, imag = 0.0):
    E251: return magic(r = real, i = imag)
    """
    parens = 0
    no_space = False
    prev_end = None
    for token_type, text, start, end, line in tokens:
        if no_space:
            no_space = False
            if start != prev_end:
                yield (prev_end,
                       "E251 no spaces around keyword / parameter equals")
        elif token_type == tokenize.OP:
            if text == '(':
                parens += 1
            elif text == ')':
                parens -= 1
            elif parens and text == '=':
                no_space = True
                if start != prev_end:
                    yield (prev_end,
                           "E251 no spaces around keyword / parameter equals")
        prev_end = end


def whitespace_before_inline_comment(logical_line, tokens):
    """
    Separate inline comments by at least two spaces.

    An inline comment is a comment on the same line as a statement.  Inline
    comments should be separated by at least two spaces from the statement.
    They should start with a # and a single space.

    Okay: x = x + 1  # Increment x
    Okay: x = x + 1    # Increment x
    E261: x = x + 1 # Increment x
    E262: x = x + 1  #Increment x
    E262: x = x + 1  #  Increment x
    """
    prev_end = (0, 0)
    for token_type, text, start, end, line in tokens:
        if token_type == tokenize.COMMENT:
            if not line[:start[1]].strip():
                continue
            if prev_end[0] == start[0] and start[1] < prev_end[1] + 2:
                yield (prev_end,
                       "E261 at least two spaces before inline comment")
            if text.startswith('#  ') or not text.startswith('# '):
                yield start, "E262 inline comment should start with '# '"
        elif token_type != tokenize.NL:
            prev_end = end


def imports_on_separate_lines(logical_line):
    r"""
    Imports should usually be on separate lines.

    Okay: import os\nimport sys
    E401: import sys, os

    Okay: from subprocess import Popen, PIPE
    Okay: from myclas import MyClass
    Okay: from foo.bar.yourclass import YourClass
    Okay: import myclass
    Okay: import foo.bar.yourclass
    """
    line = logical_line
    if line.startswith('import '):
        found = line.find(',')
        if -1 < found and ';' not in line[:found]:
            yield found, "E401 multiple imports on one line"


def compound_statements(logical_line):
    r"""
    Compound statements (multiple statements on the same line) are
    generally discouraged.

    While sometimes it's okay to put an if/for/while with a small body
    on the same line, never do this for multi-clause statements. Also
    avoid folding such long lines!

    Okay: if foo == 'blah':\n    do_blah_thing()
    Okay: do_one()
    Okay: do_two()
    Okay: do_three()

    E701: if foo == 'blah': do_blah_thing()
    E701: for x in lst: total += x
    E701: while t < 10: t = delay()
    E701: if foo == 'blah': do_blah_thing()
    E701: else: do_non_blah_thing()
    E701: try: something()
    E701: finally: cleanup()
    E701: if foo == 'blah': one(); two(); three()

    E702: do_one(); do_two(); do_three()
    """
    line = logical_line
    found = line.find(':')
    if -1 < found < len(line) - 1:
        before = line[:found]
        if (before.count('{') <= before.count('}') and  # {'a': 1} (dict)
            before.count('[') <= before.count(']') and  # [1:2] (slice)
            before.count('(') <= before.count(')') and  # (Python 3 annotation)
                not LAMBDA_REGEX.search(before)):       # lambda x: x
            yield found, "E701 multiple statements on one line (colon)"
    found = line.find(';')
    if -1 < found:
        yield found, "E702 multiple statements on one line (semicolon)"


def explicit_line_join(logical_line, tokens):
    r"""
    Avoid explicit line join between brackets.

    The preferred way of wrapping long lines is by using Python's implied line
    continuation inside parentheses, brackets and braces.  Long lines can be
    broken over multiple lines by wrapping expressions in parentheses.  These
    should be used in preference to using a backslash for line continuation.

    E502: aaa = [123, \\n       123]
    E502: aaa = ("bbb " \\n       "ccc")

    Okay: aaa = [123,\n       123]
    Okay: aaa = ("bbb "\n       "ccc")
    Okay: aaa = "bbb " \\n    "ccc"
    """
    prev_start = prev_end = parens = 0
    for token_type, text, start, end, line in tokens:
        if start[0] != prev_start and parens and backslash:
            yield backslash, "E502 the backslash is redundant between brackets"
        if end[0] != prev_end:
            if line.rstrip('\r\n').endswith('\\'):
                backslash = (end[0], len(line.splitlines()[-1]) - 1)
            else:
                backslash = None
            prev_start = prev_end = end[0]
        else:
            prev_start = start[0]
        if token_type == tokenize.OP:
            if text in '([{':
                parens += 1
            elif text in ')]}':
                parens -= 1


def comparison_to_singleton(logical_line):
    """
    Comparisons to singletons like None should always be done
    with "is" or "is not", never the equality operators.

    Okay: if arg is not None:
    E711: if arg != None:
    E712: if arg == True:

    Also, beware of writing if x when you really mean if x is not None --
    e.g. when testing whether a variable or argument that defaults to None was
    set to some other value.  The other value might have a type (such as a
    container) that could be false in a boolean context!
    """
    match = COMPARE_SINGLETON_REGEX.search(logical_line)
    if match:
        same = (match.group(1) == '==')
        singleton = match.group(2)
        msg = "'if cond is %s:'" % (('' if same else 'not ') + singleton)
        if singleton in ('None',):
            code = 'E711'
        else:
            code = 'E712'
            nonzero = ((singleton == 'True' and same) or
                       (singleton == 'False' and not same))
            msg += " or 'if %scond:'" % ('' if nonzero else 'not ')
        yield match.start(1), ("%s comparison to %s should be %s" %
                               (code, singleton, msg))


def comparison_type(logical_line):
    """
    Object type comparisons should always use isinstance() instead of
    comparing types directly.

    Okay: if isinstance(obj, int):
    E721: if type(obj) is type(1):

    When checking if an object is a string, keep in mind that it might be a
    unicode string too! In Python 2.3, str and unicode have a common base
    class, basestring, so you can do:

    Okay: if isinstance(obj, basestring):
    Okay: if type(a1) is type(b1):
    """
    match = COMPARE_TYPE_REGEX.search(logical_line)
    if match:
        inst = match.group(3)
        if inst and isidentifier(inst) and inst not in SINGLETONS:
            return  # Allow comparison for types which are not obvious
        yield match.start(1), "E721 do not compare types, use 'isinstance()'"


def python_3000_has_key(logical_line):
    r"""
    The {}.has_key() method will be removed in the future version of
    Python. Use the 'in' operation instead.

    Okay: if "alph" in d:\n    print d["alph"]
    W601: assert d.has_key('alph')
    """
    pos = logical_line.find('.has_key(')
    if pos > -1:
        yield pos, "W601 .has_key() is deprecated, use 'in'"


def python_3000_raise_comma(logical_line):
    """
    When raising an exception, use "raise ValueError('message')"
    instead of the older form "raise ValueError, 'message'".

    The paren-using form is preferred because when the exception arguments
    are long or include string formatting, you don't need to use line
    continuation characters thanks to the containing parentheses.  The older
    form will be removed in Python 3000.

    Okay: raise DummyError("Message")
    W602: raise DummyError, "Message"
    """
    match = RAISE_COMMA_REGEX.match(logical_line)
    if match and not RERAISE_COMMA_REGEX.match(logical_line):
        yield match.start(1), "W602 deprecated form of raising exception"


def python_3000_not_equal(logical_line):
    """
    != can also be written <>, but this is an obsolete usage kept for
    backwards compatibility only. New code should always use !=.
    The older syntax is removed in Python 3000.

    Okay: if a != 'no':
    W603: if a <> 'no':
    """
    pos = logical_line.find('<>')
    if pos > -1:
        yield pos, "W603 '<>' is deprecated, use '!='"


def python_3000_backticks(logical_line):
    """
    Backticks are removed in Python 3000.
    Use repr() instead.

    Okay: val = repr(1 + 2)
    W604: val = `1 + 2`
    """
    pos = logical_line.find('`')
    if pos > -1:
        yield pos, "W604 backticks are deprecated, use 'repr()'"


##############################################################################
# Helper functions
##############################################################################

if '' == ''.encode():
    # Python 2: implicit encoding.
    isidentifier = re.compile(r'[a-zA-Z_]\w*').match
else:
    # Python 3
    isidentifier = str.isidentifier


def expand_indent(line):
    r"""
    Return the amount of indentation.
    Tabs are expanded to the next multiple of 8.

    >>> expand_indent('    ')
    4
    >>> expand_indent('\t')
    8
    >>> expand_indent('    \t')
    8
    >>> expand_indent('       \t')
    8
    >>> expand_indent('        \t')
    16
    """
    if '\t' not in line:
        return len(line) - len(line.lstrip())
    result = 0
    for char in line:
        if char == '\t':
            result = result // 8 * 8 + 8
        elif char == ' ':
            result += 1
        else:
            break
    return result


def mute_string(text):
    """
    Replace contents with 'xxx' to prevent syntax matching.

    >>> mute_string('"abc"')
    '"xxx"'
    >>> mute_string("'''abc'''")
    "'''xxx'''"
    >>> mute_string("r'abc'")
    "r'xxx'"
    """
    # String modifiers (e.g. u or r)
    start = text.index(text[-1]) + 1
    end = len(text) - 1
    # Triple quotes
    if text[-3:] in ('"""', "'''"):  # '''
        start += 2
        end -= 2
    return text[:start] + 'x' * (end - start) + text[end:]


##############################################################################
# Framework to run all checks
##############################################################################


def find_checks(argument_name):
    """
    Find all globally visible functions where the first argument name
    starts with argument_name.
    """
    checks = []
    for name, function in list(globals().items()):
        if not inspect.isfunction(function):
            continue
        args = inspect.getargspec(function)[0]
        if args and args[0].startswith(argument_name):
            codes = ERRORCODE_REGEX.findall(function.__doc__ or '')
#            yield name, codes, function, args
            for code in codes or ['']:
                    if not code or not ignore_code(code):
                        checks.append((name, function, args))
                        break
    checks.sort()
    return checks


def ignore_code(code):
    """
    Check if options.ignore contains a prefix of the error code.
    If options.select contains a prefix of the error code, do not ignore it.
    """
    for select in options.select:
        if code.startswith(select):
            return False
    for ignore in options.ignore:
        if code.startswith(ignore):
            return True


#Set Physical and Logical check functions
options.physical_checks = find_checks('physical_line')
options.logical_checks = find_checks('logical_line')


def refresh_checks():
    #Refresh the Physical and Logical check functions
    options.physical_checks = find_checks('physical_line')
    options.logical_checks = find_checks('logical_line')


class Checker(object):
    """
    Load a Python source file, tokenize it, check coding style.
    """

    def __init__(self, filename, lines=None):
        self._physical_checks = options.physical_checks
        self._logical_checks = options.logical_checks
        self.max_line_length = options.max_line_length
        self.verbose = options.verbose
        self.filename = filename
        self.lines = lines
        self.results = []

    def readline(self):
        """
        Get the next line from the input buffer.
        """
        self.line_number += 1
        if self.line_number > len(self.lines):
            return ''
        return self.lines[self.line_number - 1]

    def readline_check_physical(self):
        """
        Check and return the next physical line. This method can be
        used to feed tokenize.generate_tokens.
        """
        line = self.readline()
        if line:
            self.check_physical(line)
        return line

    def run_check(self, check, argument_names):
        """
        Run a check plugin.
        """
        arguments = []
        for name in argument_names:
            arguments.append(getattr(self, name))
        return check(*arguments)

    def check_physical(self, line):
        """
        Run all physical checks on a raw input line.
        """
        self.physical_line = line
        if self.indent_char is None and line[:1] in WHITESPACE:
            self.indent_char = line[0]
        for name, check, argument_names in self._physical_checks:
            result = self.run_check(check, argument_names)
            if result is not None:
                offset, text = result
                self.report_error(self.line_number, offset, text, check)

    def build_tokens_line(self):
        """
        Build a logical line from tokens.
        """
        self.mapping = []
        logical = []
        length = 0
        previous = None
        for token in self.tokens:
            token_type, text = token[0:2]
            if token_type in SKIP_TOKENS:
                continue
            if token_type == tokenize.STRING:
                text = mute_string(text)
            if previous:
                end_row, end = previous[3]
                start_row, start = token[2]
                if end_row != start_row:    # different row
                    prev_text = self.lines[end_row - 1][end - 1]
                    if prev_text == ',' or (prev_text not in '{[('
                                            and text not in '}])'):
                        logical.append(' ')
                        length += 1
                elif end != start:  # different column
                    fill = self.lines[end_row - 1][end:start]
                    logical.append(fill)
                    length += len(fill)
            self.mapping.append((length, token))
            logical.append(text)
            length += len(text)
            previous = token
        self.logical_line = ''.join(logical)
        assert self.logical_line.strip() == self.logical_line

    def check_logical(self):
        """
        Build a line from tokens and run all logical checks on it.
        """
        self.build_tokens_line()
        first_line = self.lines[self.mapping[0][1][2][0] - 1]
        indent = first_line[:self.mapping[0][1][2][1]]
        self.previous_indent_level = self.indent_level
        self.indent_level = expand_indent(indent)
        for name, check, argument_names in self._logical_checks:
            for result in self.run_check(check, argument_names):
                offset, text = result
                if isinstance(offset, tuple):
                    orig_number, orig_offset = offset
                else:
                    for token_offset, token in self.mapping:
                        if offset >= token_offset:
                            orig_number = token[2][0]
                            orig_offset = (token[2][1] + offset - token_offset)
                self.report_error(orig_number, orig_offset, text, check)
        self.previous_logical = self.logical_line

    def generate_tokens(self):
        tokengen = tokenize.generate_tokens(self.readline_check_physical)
        try:
            for token in tokengen:
                yield token
        except (SyntaxError, tokenize.TokenError):
            exc_type, exc = sys.exc_info()[:2]
            offset = exc.args[1]
            if len(offset) > 2:
                offset = offset[1:3]
            self.report_error(offset[0], offset[1],
                              'E901 %s: %s' % (exc_type.__name__, exc.args[0]),
                              self.generate_tokens)
    generate_tokens.__doc__ = "    Check if the syntax is valid."

    def check_all(self, expected=None, line_offset=0):
        """
        Run all checks on the input file.
        """
        self.line_number = 0
        self.line_offset = line_offset
        self.indent_char = None
        self.indent_level = 0
        self.previous_logical = ''
        self.tokens = []
        self.blank_lines = blank_lines_before_comment = 0
        parens = 0
        for token in self.generate_tokens():
            self.tokens.append(token)
            token_type, text = token[0:2]
            if self.verbose >= 3:
                if token[2][0] == token[3][0]:
                    pos = '[%s:%s]' % (token[2][1] or '', token[3][1])
                else:
                    pos = 'l.%s' % token[3][0]
            if token_type == tokenize.OP:
                if text in '([{':
                    parens += 1
                elif text in '}])':
                    parens -= 1
            elif not parens:
                if token_type == tokenize.NEWLINE:
                    if self.blank_lines < blank_lines_before_comment:
                        self.blank_lines = blank_lines_before_comment
                    self.check_logical()
                    self.tokens = []
                    self.blank_lines = blank_lines_before_comment = 0
                elif token_type == tokenize.NL:
                    if len(self.tokens) == 1:
                        # The physical line contains only this token.
                        self.blank_lines += 1
                    self.tokens = []
                elif token_type == tokenize.COMMENT and len(self.tokens) == 1:
                    if blank_lines_before_comment < self.blank_lines:
                        blank_lines_before_comment = self.blank_lines
                    self.blank_lines = 0
                    if COMMENT_WITH_NL:
                        # The comment also ends a physical line
                        self.tokens = []
        return self.results

    def report_error(self, line_number, offset, text, check):
        """
        Report an error, according to options.
        """
        if options.repeat:
            self.results.append("%s:%s:%d: %s" %
                    (self.filename, self.line_offset + line_number,
                     offset + 1, text))
            if options.show_source and len(self.lines) >= line_number:
                line = self.lines[line_number - 1]
                self.results.append(line.rstrip())
                self.results.append(' ' * offset + '^')
            if options.show_pep8:
                self.results.append(check.__doc__.lstrip('\n').rstrip())


def run_check(fileName, source):
    """
    Parse options and run checks on Python source.
    """
    try:
        lines = ['%s\n' % line for line in source.splitlines()]
        return Checker(fileName, lines).check_all()
    except Exception as reason:
        print(('pep8mod couldn\'t parse file: {0}'.format(fileName)))
        print(reason)
        raise
    return []



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt4.QtGui import QTreeWidget
from PyQt4.QtGui import QTreeWidgetItem
from PyQt4.QtGui import QIcon
from PyQt4.QtGui import QAbstractItemView
from PyQt4.QtGui import QHeaderView
from PyQt4.QtGui import QCursor
from PyQt4.QtGui import QMenu

from PyQt4.QtCore import Qt
from PyQt4.QtCore import SIGNAL

from ninja_ide import resources


class TreeSymbolsWidget(QTreeWidget):

###############################################################################
# TreeSymbolsWidget SIGNALS
###############################################################################

    """
    goToDefinition(int)
    """

###############################################################################

    def __init__(self):
        QTreeWidget.__init__(self)
        self.header().setHidden(True)
        self.setSelectionMode(self.SingleSelection)
        self.setAnimated(True)
        self.header().setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.header().setResizeMode(0, QHeaderView.ResizeToContents)
        self.header().setStretchLastSection(False)
        self.actualSymbols = ('', {})
        self.docstrings = {}
        self.collapsedItems = {}

        self.connect(self, SIGNAL("itemClicked(QTreeWidgetItem *, int)"),
                     self._go_to_definition)
        self.connect(self, SIGNAL("itemActivated(QTreeWidgetItem *, int)"),
                     self._go_to_definition)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.connect(self,
                     SIGNAL("customContextMenuRequested(const QPoint &)"),
                     self._menu_context_tree)
        self.connect(self, SIGNAL("itemCollapsed(QTreeWidgetItem *)"),
                     self._item_collapsed)
        self.connect(self, SIGNAL("itemExpanded(QTreeWidgetItem *)"),
                     self._item_expanded)

    def select_current_item(self, line, col):
        #TODO
        #print line, col
        pass

    def _menu_context_tree(self, point):
        index = self.indexAt(point)
        if not index.isValid():
            return

        menu = QMenu(self)
        f_all = menu.addAction(self.tr("Fold all"))
        u_all = menu.addAction(self.tr("Unfold all"))
        menu.addSeparator()
        u_class = menu.addAction(self.tr("Unfold classes"))
        u_class_method = menu.addAction(self.tr("Unfold classes and methods"))
        u_class_attr = menu.addAction(self.tr("Unfold classes and attributes"))
        menu.addSeparator()
        #save_state = menu.addAction(self.tr("Save State"))

        self.connect(f_all, SIGNAL("triggered()"),
                     lambda: self.collapseAll())
        self.connect(u_all, SIGNAL("triggered()"),
                     lambda: self.expandAll())
        self.connect(u_class, SIGNAL("triggered()"), self._unfold_class)
        self.connect(u_class_method, SIGNAL("triggered()"),
                     self._unfold_class_method)
        self.connect(u_class_attr, SIGNAL("triggered()"),
                     self._unfold_class_attribute)
        #self.connect(save_state, SIGNAL("triggered()"),
            #self._save_symbols_state)

        menu.exec_(QCursor.pos())

    def _get_classes_root(self):
        class_root = None
        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            if item.isClass and not item.isClickable:
                class_root = item
                break
        return class_root

    def _unfold_class(self):
        self.collapseAll()
        classes_root = self._get_classes_root()
        if not classes_root:
            return

        classes_root.setExpanded(True)

    def _unfold_class_method(self):
        self.expandAll()
        classes_root = self._get_classes_root()
        if not classes_root:
            return
        #for each class!
        for i in range(classes_root.childCount()):
            class_item = classes_root.child(i)
            #for each attribute or functions
            for j in range(class_item.childCount()):
                item = class_item.child(j)
                #METHODS ROOT!!
                if not item.isMethod and not item.isClickable:
                    item.setExpanded(False)
                    break

    def _unfold_class_attribute(self):
        self.expandAll()
        classes_root = self._get_classes_root()
        if not classes_root:
            return
        #for each class!
        for i in range(classes_root.childCount()):
            class_item = classes_root.child(i)
            #for each attribute or functions
            for j in range(class_item.childCount()):
                item = class_item.child(j)
                #ATTRIBUTES ROOT!!
                if not item.isAttribute and not item.isClickable:
                    item.setExpanded(False)
                    break

    def _save_symbols_state(self):
        #filename = self.actualSymbols[0]
        #TODO: persist self.collapsedItems[filename] in QSettings
        pass

    def _get_expand(self, item):
        """
        Returns True or False to be used as setExpanded() with the items
        It method is based on the click that the user made in the tree
        """
        name = self._get_unique_name(item)
        filename = self.actualSymbols[0]
        collapsed_items = self.collapsedItems.get(filename, [])
        can_check = (not item.isClickable) or item.isClass or item.isMethod
        if can_check and name in collapsed_items:
            return False
        return True

    @staticmethod
    def _get_unique_name(item):
        """
        Returns a string used as unique name
        """
        # className_Attributes/className_Functions
        parent = item.parent()
        if parent:
            return "%s_%s" % (parent.text(0), item.text(0))
        return "_%s" % item.text(0)

    def update_symbols_tree(self, symbols, filename='', parent=None):
        if not parent:
            if filename == self.actualSymbols[0] and \
                    self.actualSymbols[1] and not symbols:
                    return

            if symbols == self.actualSymbols[1]:
                # Nothing new then return
                return

            # we have new symbols refresh it
            self.clear()
            self.actualSymbols = (filename, symbols)
            self.docstrings = symbols.get('docstrings', {})
            parent = self
        if 'attributes' in symbols:
            globalAttribute = ItemTree(parent, [self.tr("Attributes")])
            globalAttribute.isClickable = False
            globalAttribute.isAttribute = True
            globalAttribute.setExpanded(self._get_expand(globalAttribute))
            for glob in sorted(symbols['attributes']):
                globItem = ItemTree(globalAttribute, [glob],
                                    lineno=symbols['attributes'][glob])
                globItem.isAttribute = True
                globItem.setIcon(0, QIcon(resources.IMAGES['attribute']))
                globItem.setExpanded(self._get_expand(globItem))

        if 'functions' in symbols and symbols['functions']:
            functionsItem = ItemTree(parent, [self.tr("Functions")])
            functionsItem.isClickable = False
            functionsItem.isMethod = True
            functionsItem.setExpanded(self._get_expand(functionsItem))
            for func in sorted(symbols['functions']):
                item = ItemTree(functionsItem, [func],
                                lineno=symbols['functions'][func]['lineno'])
                tooltip = self.create_tooltip(
                    func, symbols['functions'][func]['lineno'])
                item.isMethod = True
                item.setIcon(0, QIcon(resources.IMAGES['function']))
                item.setToolTip(0, tooltip)
                item.setExpanded(self._get_expand(item))
                self.update_symbols_tree(
                    symbols['functions'][func]['functions'], parent=item)
        if 'classes' in symbols and symbols['classes']:
            classItem = ItemTree(parent, [self.tr("Classes")])
            classItem.isClickable = False
            classItem.isClass = True
            classItem.setExpanded(self._get_expand(classItem))
            for claz in sorted(symbols['classes']):
                line_number = symbols['classes'][claz]['lineno']
                item = ItemTree(classItem, [claz], lineno=line_number)
                item.isClass = True
                tooltip = self.create_tooltip(claz, line_number)
                item.setToolTip(0, tooltip)
                item.setIcon(0, QIcon(resources.IMAGES['class']))
                item.setExpanded(self._get_expand(item))
                self.update_symbols_tree(symbols['classes'][claz]['members'],
                                         parent=item)

    def _go_to_definition(self, item):
        if item.isClickable:
            self.emit(SIGNAL("goToDefinition(int)"), item.lineno - 1)

    def create_tooltip(self, name, lineno):
        doc = self.docstrings.get(lineno, None)
        if doc is None:
            doc = ''
        else:
            doc = '\n' + doc
        tooltip = name + doc
        return tooltip

    def _item_collapsed(self, item):
        super(TreeSymbolsWidget, self).collapseItem(item)

        can_check = (not item.isClickable) or item.isClass or item.isMethod
        if can_check:
            n = self._get_unique_name(item)
            filename = self.actualSymbols[0]
            self.collapsedItems.setdefault(filename, [])
            if not n in self.collapsedItems[filename]:
                self.collapsedItems[filename].append(n)

    def _item_expanded(self, item):
        super(TreeSymbolsWidget, self).expandItem(item)

        n = self._get_unique_name(item)
        filename = self.actualSymbols[0]
        if n in self.collapsedItems.get(filename, []):
            self.collapsedItems[filename].remove(n)
            if not len(self.collapsedItems[filename]):
                # no more items, free space
                del self.collapsedItems[filename]

    def clean(self):
        """
        Reset the tree and reset attributes
        """
        self.clear()
        self.collapsedItems = {}


class ItemTree(QTreeWidgetItem):

    def __init__(self, parent, name, lineno=None):
        QTreeWidgetItem.__init__(self, parent, name)
        self.lineno = lineno
        self.isClickable = True
        self.isAttribute = False
        self.isClass = False
        self.isMethod = False



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import sys

try:
    #if sys.platform == 'win32':
        #from ninja_ide.core.filesystem_notifications import windows
        #source = windows
    #elif sys.platform == 'darwin':
    from ninja_ide.core.filesystem_notifications import darwin
    source = darwin
    #elif sys.platform.startswith("linux"):
        #from ninja_ide.core.filesystem_notifications import linux
        #source = linux
    #else:
        ##Aything we do not have a clue how to handle
        #from ninja_ide.core.filesystem_notifications import openbsd
        #source = openbsd
except:
    pass
    #from ninja_ide.core.filesystem_notifications import openbsd
    #source = openbsd


NinjaFileSystemWatcher = source.NinjaFileSystemWatcher()



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

import os
import sys
import shutil
import copy
import zipfile
import traceback
#lint:disable
try:
    from urllib.request import urlopen
    from urllib.error import URLError
except ImportError:
    from urllib2 import urlopen
    from urllib2 import URLError
#lint:enable

from ninja_ide import resources
from ninja_ide.tools.logger import NinjaLogger
from ninja_ide.tools import json_manager

logger = NinjaLogger('ninja_ide.core.plugin_manager')
REQUIREMENTS = 'requirements.txt'
COMMAND_FOR_PIP_INSTALL = 'pip install -r %s'
try:
    # For Python2
    str = unicode  # lint:ok
except NameError:
    # We are in Python3
    pass


class ServiceLocator(object):

    '''
    Hold the services and allows the interaction between NINJA-IDE and plugins
    '''

    def __init__(self, services=None):
        self.__services = services if services else {}

    def get_service(self, name):
        return self.__services.get(name)

    def get_availables_services(self):
        return list(self.__services.keys())


'''
NINJA-IDE Plugin
my_plugin.plugin

{
  "module": "my_plugin",
  "class": "MyPluginExample",
  "authors": "Martin Alderete <malderete@gmail.com>",
  "version": "0.1",
  "description": "Este plugin es de prueba"
}

class MyPluginExample(Plugin):

    def initialize(self):
        #Configure the plugin using the NINJA-IDE API!!!
        self.editor_s = self.service_locator.get_service('editor')
        self.toolbar_s = self.service_locator.get_service('toolbar')
        self.toolbar_s.add_action(QAction(...........))
        self.appmenu_s = self.service_locator.get_service('appmenu')
        self.appmenu_s.add_menu(QMenu(......))
        #connect events!
        self.editor_s.editorKeyPressEvent.connect(self.my_plugin_key_pressed)

    def my_plugin_key_pressed(self, ...):
        print 'se apreto alguna tecla en el ide...'

'''
###############################################################################
# NINJA-IDE Plugin Manager
###############################################################################


class PluginManagerException(Exception):
    pass


#Singleton
__pluginManagerInstance = None


def PluginManager(*args, **kw):
    global __pluginManagerInstance
    if __pluginManagerInstance is None:
        __pluginManagerInstance = __PluginManager(*args, **kw)
    return __pluginManagerInstance

#Extension of the NINJA-IDE plugin
PLUGIN_EXTENSION = '.plugin'


class __PluginManager(object):
    '''
    Plugin manager allows to load, unload, initialize plugins.
    '''

    def __init__(self, plugins_dir, service_locator):
        '''
        @param plugins_dir: Path to search plugins.
        @param service_loctor: ServiceLocator object.
        '''
        self._service_locator = service_locator
        #new!
        self._plugins_by_dir = {}
        #add all the plugins paths
        for path in self.__create_list(plugins_dir):
            self.add_plugin_dir(path)
        #end new!
        #self._plugins_dir = plugins_dir
        self._errors = []
        #found plugins
        #example: ["logger", "my_plugin"]
        self._found_plugins = []
        #active plugins
        #example: {"logger": (LoggerIntance, metadata),
        #    "my_plugin": (MyPluginInstance, metadata)}
        self._active_plugins = {}

    def __create_list(self, obj):
        if isinstance(obj, (list, tuple)):
            return obj
        #string then returns a list of one item!
        return [obj]

    def add_plugin_dir(self, plugin_dir):
        '''
        Add a new directory to search plugins.

        @param plugin_dir: absolute path.
        '''
        if not plugin_dir in self._plugins_by_dir:
            self._plugins_by_dir[plugin_dir] = []

    def get_actives_plugins(self):
        import warnings
        warnings.warn("Deprecated in behalf of a TYPO free method name")
        return self.get_active_plugins()

    def get_active_plugins(self):
        '''
        Returns a list the instances
        '''
        return [plugin[0] for plugin in list(self._active_plugins.values())]

    def _get_dir_from_plugin_name(self, plugin_name):
        '''
        Returns the dir of the plugin_name
        '''
        for dir_, plug_names in list(self._plugins_by_dir.items()):
            if plugin_name in plug_names:
                return dir_

    def __getitem__(self, plugin_name):
        '''
        Magic method to get a plugin instance
        from a given name.
        @Note: This method has the logic below.
        Check if the plugin is known,
        if it is active return it,
        otherwise, active it and return it.
        If the plugin name does not exist
        raise KeyError exception.

        @param plugin_name: plugin name.

        @return: Plugin instance or None
        '''
        global PLUGIN_EXTENSION
        ext = PLUGIN_EXTENSION
        if not plugin_name.endswith(ext):
            plugin_name += ext

        if plugin_name in self._found_plugins:
            if not plugin_name in self._active_plugins:
                dir_ = self._get_dir_from_plugin_name(plugin_name)
                self.load(plugin_name, dir_)
            return self._active_plugins[plugin_name][0]
        raise KeyError(plugin_name)

    def __contains__(self, plugin_name):
        '''
        Magic method to know whether the
        PluginManager contains
        a plugin with a given name.

        @param plugin_name: plugin name.

        @return: True or False.
        '''
        return plugin_name in self._found_plugins

    def __iter__(self):
        '''
        Magic method to iterate over all
        the plugin's names.

        @return: iterator.
        '''
        return iter(self._found_plugins)

    def __len__(self):
        '''
        Magic method to know the plugins
        quantity.
        @return: length.
        '''
        return len(self._found_plugins)

    def __bool__(self):
        '''
        Magic method to indicate that any
        instance must pass the if conditional
        if x:
        '''
        return True

    def get_plugin_name(self, file_name):
        '''
        Get the plugin's name from a file name.
        @param file_name: A file object name.
        @return: A plugin name from a file.
        '''
        plugin_file_name, file_ext = os.path.splitext(file_name)
        return plugin_file_name

    def list_plugins(self, dir_name):
        '''
        Crawl a directory and collect plugins.
        @return: List with plugin names.
        '''
        global PLUGIN_EXTENSION
        ext = PLUGIN_EXTENSION
        try:
            listdir = os.listdir(dir_name)
            return [plug for plug in listdir if plug.endswith(ext)]
        except OSError:
            return ()

    def is_plugin_active(self, plugin_name):
        '''
        Check if a plugin is or not active
        @param plugin_name: Plugin name to check.
        @return: True or False
        '''
        return plugin_name in self._active_plugins

    def discover(self):
        '''
        Search all files for directory
        and get the valid plugin's names.
        '''
        for dir_name in self._plugins_by_dir:
            for file_name in self.list_plugins(dir_name):
                plugin_name = file_name
                if not plugin_name in self._found_plugins:
                    self._found_plugins.append(plugin_name)
                    self._plugins_by_dir[dir_name].append(plugin_name)

    def _load_module(self, module, klassname, metadata, dir_name):
        old_syspath = copy.copy(sys.path)
        try:
            sys.path.insert(1, dir_name)
            module = __import__(module, globals(), locals(), [])
            klass = getattr(module, klassname)
            #Instanciate the plugin
            plugin_instance = klass(self._service_locator, metadata=metadata)
            #return the plugin instance
            return plugin_instance
        except(ImportError, AttributeError) as reason:
            raise PluginManagerException('Error loading "%s": %s' %
                                         (module, reason))
        finally:
            sys.path = old_syspath
        return None

    def load(self, plugin_name, dir_name):
        global PLUGIN_EXTENSION
        if plugin_name in self._active_plugins:
            return
        for dir_name, plugin_list in list(self._plugins_by_dir.items()):
            if plugin_name in plugin_list:
                ext = PLUGIN_EXTENSION
                plugin_filename = os.path.join(dir_name, plugin_name)
                plugin_structure = json_manager.read_json(plugin_filename)
                plugin_structure['name'] = plugin_name.replace(ext, '')
                module = plugin_structure.get('module', None)
                klassname = plugin_structure.get('class', None)
                if module is not None and klassname is not None:
                    try:
                        plugin_instance = self._load_module(module,
                                                            klassname,
                                                            plugin_structure,
                                                            dir_name)
                        #set a get_plugin method to get the reference to other
                        #setattr(plugin_instance,'get_plugin',self.__getitem__)
                        #call a special method *initialize* in the plugin!
                        plugin_instance.metadata = plugin_structure
                        logger.info("Calling initialize (%s)", plugin_name)
                        plugin_instance.initialize()
                        #tuple (instance, metadata)
                        plugin_metadata = (plugin_instance, plugin_structure)
                        self._active_plugins[plugin_name] = plugin_metadata
                    except (PluginManagerException, Exception) as reason:
                        logger.error("Not instanciated (%s): %s", plugin_name,
                                     reason)
                        #remove the plugin because has errors
                        self._found_plugins.remove(plugin_name)
                        traceback_msg = traceback.format_exc()
                        plugin_name = plugin_name.replace(ext, '')
                        #add the traceback to errors
                        self._add_error(plugin_name, traceback_msg)
                    else:
                        logger.info("Successfuly initialized (%s)",
                                    plugin_name)

    def load_all(self):
        for dir, pl in list(self._plugins_by_dir.items()):
            #Copy the list because may be we REMOVE item while iterate!
            found_plugins_aux = copy.copy(pl)
            for plugin_name in found_plugins_aux:
                self.load(plugin_name, dir)

    def load_all_external(self, plugin_path):
        #Copy the list because may be we REMOVE item while iterate!
        found_plugins_aux = copy.copy(self._found_plugins)
        for plugin_name in found_plugins_aux:
            self.load(plugin_name, plugin_path)

    def unload(self, plugin_name):
        try:
            plugin_object = self._active_plugins[plugin_name][0]
            #call a special method *finish* in the plugin!
            plugin_object.finish()
            del self._active_plugins[plugin_name]
        except Exception as reason:
            logger.error("Finishing plugin (%s): %s", plugin_name, reason)
        else:
            logger.info("Successfuly finished (%s)", plugin_name)

    def unload_all(self):
        #Copy the list because may be we REMOVE item while iterate!
        active_plugins_aux = copy.copy(self._active_plugins)
        for plugin_name in active_plugins_aux:
            self.unload(plugin_name)

    def shutdown(self):
        self.unload_all()

    def get_availables_services(self):
        """
        Returns all services availables
        """
        self._service_locator.get_availables_services()

    def _add_error(self, plugin_name, traceback_msg):
        self._errors.append((plugin_name, traceback_msg))

    @property
    def errors(self):
        """
        Returns a comma separated values of errors
        """
        return self._errors


def _availables_plugins(url):
    """
    Return the availables plugins from an url in NINJA-IDE web page
    """
    try:
        descriptor = urlopen(url)
        plugins = json_manager.read_json_from_stream(descriptor)
        return plugins
    except URLError:
        return {}


def available_oficial_plugins():
    '''
    Returns a dict with OFICIAL availables plugins in NINJA-IDE web page
    '''
    return _availables_plugins(resources.PLUGINS_WEB)


def available_community_plugins():
    '''
    Returns a dict with COMMUNITY availables plugins in NINJA-IDE web page
    '''
    return _availables_plugins(resources.PLUGINS_COMMUNITY)


def local_plugins():
    '''
    Returns the local plugins
    '''
    if not os.path.isfile(resources.PLUGINS_DESCRIPTOR):
        return []
    plugins = json_manager.read_json(resources.PLUGINS_DESCRIPTOR)
    return plugins


def __get_all_plugin_descriptors():
    '''
    Returns all the .plugin files
    '''
    global PLUGIN_EXTENSION
    return [pf for pf in os.listdir(resources.PLUGINS)
            if pf.endswith(PLUGIN_EXTENSION)]


def download_plugin(file_):
    '''
    Download a plugin specified by file_
    '''
    global PLUGIN_EXTENSION
    #get all the .plugin files in local filesystem
    plugins_installed_before = set(__get_all_plugin_descriptors())
    #download the plugin
    fileName = os.path.join(resources.PLUGINS, os.path.basename(file_))
    content = urlopen(file_)
    f = open(fileName, 'wb')
    f.write(content.read())
    f.close()
    #create the zip
    zipFile = zipfile.ZipFile(fileName, 'r')
    zipFile.extractall(resources.PLUGINS)
    zipFile.close()
    #clean up the enviroment
    os.remove(fileName)
    #get the name of the last installed plugin
    plugins_installed_after = set(__get_all_plugin_descriptors())
    #using set operations get the difference that is the new plugin
    new_plugin = (plugins_installed_after - plugins_installed_before).pop()
    return new_plugin


def manual_install(file_):
    """Copy zip file and install."""
    global PLUGIN_EXTENSION
    #get all the .plugin files in local filesystem
    plugins_installed_before = set(__get_all_plugin_descriptors())
    #copy the plugin
    fileName = os.path.join(resources.PLUGINS, os.path.basename(file_))
    shutil.copyfile(file_, fileName)
    #extract the zip
    zipFile = zipfile.ZipFile(fileName, 'r')
    zipFile.extractall(resources.PLUGINS)
    zipFile.close()
    #clean up the enviroment
    os.remove(fileName)
    #get the name of the last installed plugin
    plugins_installed_after = set(__get_all_plugin_descriptors())
    #using set operations get the difference that is the new plugin
    new_plugin = (plugins_installed_after - plugins_installed_before).pop()
    return new_plugin


def has_dependencies(plug):
    global REQUIREMENTS, COMMAND_FOR_PIP_INSTALL

    plugin_name = plug[0]
    structure = []
    if os.path.isfile(resources.PLUGINS_DESCRIPTOR):
        structure = json_manager.read_json(resources.PLUGINS_DESCRIPTOR)
    PLUGINS = resources.PLUGINS
    for p in structure:
        if p['name'] == plugin_name:
            pd_file = os.path.join(PLUGINS, p['plugin-descriptor'])
            p_json = json_manager.read_json(pd_file)
            module = p_json.get('module')
            #plugin_module/requirements.txt
            req_file = os.path.join(os.path.join(PLUGINS, module),
                                    REQUIREMENTS)
            if os.path.isfile(req_file):
                return (True, COMMAND_FOR_PIP_INSTALL % req_file)
            #the plugin was found but no requirement then break!
            break
    return (False, None)


def update_local_plugin_descriptor(plugins):
    '''
    updates the local plugin description
    The description.json file holds the information about the plugins
    downloaded with NINJA-IDE
    This is a way to track the versions of the plugins
    '''
    structure = []
    if os.path.isfile(resources.PLUGINS_DESCRIPTOR):
        structure = json_manager.read_json(resources.PLUGINS_DESCRIPTOR)
    for plug_list in plugins:
        #create the plugin data
        plug = {}
        plug['name'] = plug_list[0]
        plug['version'] = plug_list[1]
        plug['description'] = plug_list[2]
        plug['authors'] = plug_list[3]
        plug['home'] = plug_list[4]
        plug['download'] = plug_list[5]
        plug['plugin-descriptor'] = plug_list[6]
        #append the plugin data
        structure.append(plug)
    json_manager.write_json(structure, resources.PLUGINS_DESCRIPTOR)


def uninstall_plugin(plug):
    """
    Uninstall the given plugin
    """
    plugin_name = plug[0]
    structure = []
    if os.path.isfile(resources.PLUGINS_DESCRIPTOR):
        structure = json_manager.read_json(resources.PLUGINS_DESCRIPTOR)
    #copy the strcuture we iterate and remove at the same time
    structure_aux = copy.copy(structure)
    for plugin in structure_aux:
        if plugin["name"] == plugin_name:
            fileName = plugin["plugin-descriptor"]
            structure.remove(plugin)
            break
    #open <plugin>.plugin file and get the module to remove
    fileName = os.path.join(resources.PLUGINS, fileName)
    plugin = json_manager.read_json(fileName)
    module = plugin.get('module')
    if module:
        pluginDir = os.path.join(resources.PLUGINS, module)
        folders = [pluginDir]
        for root, dirs, files in os.walk(pluginDir):
            pluginFiles = [os.path.join(root, f) for f in files]
            #remove all files
            list(map(os.remove, pluginFiles))
            #collect subfolders
            folders += [os.path.join(root, d) for d in dirs]
        folders.reverse()
        for f in folders:
            if os.path.isdir(f):
                os.removedirs(f)
        #remove ths plugin_name.plugin file
        os.remove(fileName)
    #write the new info
    json_manager.write_json(structure, resources.PLUGINS_DESCRIPTOR)


###############################################################################
# Module Test
###############################################################################

if __name__ == '__main__':
    folders = resources.PLUGINS
    services = {}
    sl = ServiceLocator(services)
    pm = PluginManager(folders, sl)
    #There are not plugins yet...lets discover
    pm.discover()
    logger.info("listing plugins names...")
    for p in pm:
        print(p)
    logger.info("Activating plugins...")
    pm.load_all()
    logger.info("Plugins already actives...")
    logger.info(pm.get_active_plugins())



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import os
import sys

from PyQt4.QtGui import QWizard
from PyQt4.QtGui import QWizardPage
from PyQt4.QtGui import QLabel
from PyQt4.QtGui import QPushButton
from PyQt4.QtGui import QComboBox
from PyQt4.QtGui import QLineEdit
from PyQt4.QtGui import QPlainTextEdit
from PyQt4.QtGui import QGridLayout
from PyQt4.QtGui import QVBoxLayout
from PyQt4.QtGui import QHBoxLayout
from PyQt4.QtGui import QMessageBox
from PyQt4.QtGui import QFileDialog
from PyQt4.QtGui import QListWidget
from PyQt4.QtGui import QCompleter
from PyQt4.QtGui import QDirModel
from PyQt4.QtGui import QPixmap
from PyQt4.QtCore import Qt
from PyQt4.QtCore import SIGNAL

from ninja_ide import resources
from ninja_ide.core import settings
from ninja_ide.core import plugin_interfaces
from ninja_ide.core import file_manager
from ninja_ide.tools import json_manager
from ninja_ide.tools.logger import NinjaLogger


logger = NinjaLogger('ninja_ide.gui.dialogs.wizard_new_project')
logger.info("loaded")
DEBUG = logger.debug


###############################################################################
# Wizard handler and Python Project handler
###############################################################################

class WizardNewProject(QWizard):
    """
    Wizard to create a new project (of any kind), it implements the base
    behavior. Also, it set two special projects type handler
    (PythonProjectHandler, ImportFromSourcesProjectHandler)
    """
    def __init__(self, parent):
        QWizard.__init__(self, parent)
        self.__explorer = parent
        self.setWindowTitle(self.tr("NINJA - New Project Wizard"))
        self.setPixmap(QWizard.LogoPixmap, QPixmap(resources.IMAGES['icon']))

        self.option = 'Python'
        #Add a project type handler for Python
        settings.set_project_type_handler(self.option,
            PythonProjectHandler(self.__explorer))
        #Add a project type handler for Import from existing sources
        settings.set_project_type_handler('Import from sources',
            ImportFromSourcesProjectHandler(self.__explorer))

        self.projectTypePage = PageProjectType(self)
        self.addPage(self.projectTypePage)
        self.addPage(PageProjectProperties())

        self.setButtonLayout([
            QWizard.BackButton,
            QWizard.Stretch,
            QWizard.NextButton,
            QWizard.FinishButton])

    def add_project_pages(self, option='Python'):
        logger.debug("Add project pages")
        self.option = option
        pages = settings.get_project_type_handler(option).get_pages()
        listIds = self.pageIds()
        listIds.pop(listIds.index(0))
        for page in pages:
            self.addPage(page)
        #this page is always needed by a project!
        self.addPage(PageProjectProperties())
        for i in listIds:
            self.removePage(i)

    def done(self, result):
        logger.debug("Done")
        if result == 1:
            page = self.currentPage()
            if type(page) == PageProjectProperties:
                venv = page.vtxtPlace.text()
                if venv:
                    if sys.platform == 'win32':
                        venv = os.path.join(venv, 'Scripts', 'python.exe')
                    else:
                        venv = os.path.join(venv, 'bin', 'python')
                    #check if venv folder exists
                    if not os.path.exists(venv):
                        btnPressed = QMessageBox.information(self,
                            self.tr("Virtualenv Folder"),
                            self.tr("Folder don't exists or this is not a "
                                "valid Folder.\n If you want to set "
                                "or modify, go to project properties"),
                            self.tr("Back"),
                            self.tr("Continue"))
                        if btnPressed == QMessageBox.NoButton:
                            return
                        else:
                            self.currentPage().vtxtPlace.setText("")
                page.vtxtPlace.setText(venv)
            settings.get_project_type_handler(self.option)\
                .on_wizard_finish(self)
        super(WizardNewProject, self).done(result)

    def _load_project(self, path):
        """
        Open Project based on path into Explorer
        """
        self.__explorer.open_project_folder(path)

    def _load_project_with_extensions(self, path, extensions):
        """Open Project based on path into Explorer with extensions"""
#        self._main._properties._treeProjects.load_project(
#            self._main.open_project_with_extensions(path), path)
        pass


###############################################################################
# Python Project Handler
###############################################################################


class PythonProjectHandler(plugin_interfaces.IProjectTypeHandler):
    """
    Handler for Python projects
    """
    def __init__(self, explorer):
        self.__explorer = explorer

    def get_pages(self):
        """
        Get extra pages for new Python projects
        """
        return ()

    def get_context_menus(self):
        """
        Get a special menu for new Python projects
        """
        return ()

    def on_wizard_finish(self, wizard):
        """
        Create the ninja_porject (.nja file), create the main __init__.py
        and open the project
        """
        ids = wizard.pageIds()
        page = wizard.page(ids[1])
        name = page.txtName.text()
        path = os.path.join(page.txtPlace.text(), name)
        if not path:
            QMessageBox.critical(self, self.tr("Incorrect Location"),
                self.tr("The project couldn\'t be create"))
            return
        if not os.path.exists(path):
            os.makedirs(path)
        project = {}
        project['name'] = name
        project['description'] = page.txtDescription.toPlainText()
        project['license'] = page.cboLicense.currentText()
        project['venv'] = page.vtxtPlace.text()
        json_manager.create_ninja_project(path, name, project)
        try:
            file_manager.create_init_file(path)
        except:
            logger.debug("The __init__ file already exists - Import Sources.")
        wizard._load_project(path)


###############################################################################
# Import project from existing sources
###############################################################################


class ImportFromSourcesProjectHandler(plugin_interfaces.IProjectTypeHandler):
    """
    Handler for Import from existing sources project
    """
    def __init__(self, explorer):
        self.__explorer = explorer

    def get_pages(self):
        """
        Get extra pages for Import from sources projects
        """
        return ()

    def get_context_menus(self):
        """
        Get a special menu for Import from sources projects
        """
        return ()

    def on_wizard_finish(self, wizard):
        """
        Create the ninja_porject (.nja file) and open the project
        """
        ids = wizard.pageIds()
        page = wizard.page(ids[1])
        name = page.txtName.text()
        path = page.txtPlace.text()
        if not path:
            QMessageBox.critical(self, self.tr("Incorrect Location"),
                self.tr("The project couldn\'t be create"))
            return
        project = {}
        project['name'] = name
        project['description'] = page.txtDescription.toPlainText()
        project['license'] = page.cboLicense.currentText()
        project['venv'] = page.vtxtPlace.text()
        project['project-type'] = wizard.option
        json_manager.create_ninja_project(path, name, project)
        wizard._load_project(path)


###############################################################################
# WIZARD FIRST PAGE
###############################################################################


class PageProjectType(QWizardPage):

    def __init__(self, wizard):
        QWizardPage.__init__(self)
        self.setTitle(self.tr("Project Type"))
        self.setSubTitle(self.tr("Choose the Project Type"))
        self._wizard = wizard

        vbox = QVBoxLayout(self)
        self.listWidget = QListWidget()
        vbox.addWidget(self.listWidget)
        types = settings.get_all_project_types()
        types.sort()
        index = types.index('Python')
        types.insert(0, types.pop(index))
        self.listWidget.addItems(types)
        self.listWidget.setCurrentRow(0)

        self.connect(self.listWidget,
            SIGNAL("itemClicked(QListWidgetItem *)"),
            self.load_pages)

    def validatePage(self):
        self._wizard.option = self.listWidget.currentItem().text()
        return True

    def load_pages(self):
        self.wizard().add_project_pages(self.listWidget.currentItem().text())


###############################################################################
# WIZARD LAST PAGE
###############################################################################


class PageProjectProperties(QWizardPage):

    def __init__(self):
        QWizardPage.__init__(self)
        self.setTitle(self.tr("New Project Data"))
        self.setSubTitle(self.tr(
            "Complete the following fields to create the Project Structure"))

        gbox = QGridLayout(self)
        #Names of the fields to complete
        self.lblName = QLabel(self.tr("New Project Name (*):"))
        self.lblPlace = QLabel(self.tr("Create in (*):"))
        self.lblDescription = QLabel(self.tr("Project Description:"))
        self.lblLicense = QLabel(self.tr("Project License:"))
        self.lblVenvFolder = QLabel(self.tr("Virtualenv Folder:"))
        gbox.addWidget(self.lblName, 0, 0, Qt.AlignRight)
        gbox.addWidget(self.lblPlace, 1, 0, Qt.AlignRight)
        gbox.addWidget(self.lblDescription, 2, 0, Qt.AlignTop)
        gbox.addWidget(self.lblLicense, 3, 0, Qt.AlignRight)
        gbox.addWidget(self.lblVenvFolder, 4, 0, Qt.AlignRight)

        #Fields on de right of the grid
        #Name
        self.txtName = QLineEdit()
        #Location
        hPlace = QHBoxLayout()
        self.txtPlace = QLineEdit()
        self.txtPlace.setReadOnly(True)
        self.btnExamine = QPushButton(self.tr("Browse..."))
        hPlace.addWidget(self.txtPlace)
        hPlace.addWidget(self.btnExamine)
        #Virtualenv
        vPlace = QHBoxLayout()
        self.vtxtPlace = QLineEdit()
        self._dir_completer = QCompleter()
        self._dir_completer.setModel(QDirModel(self._dir_completer))
        self.vtxtPlace.setCompleter(self._dir_completer)
        self.vbtnExamine = QPushButton(self.tr("Browse..."))
        vPlace.addWidget(self.vtxtPlace)
        vPlace.addWidget(self.vbtnExamine)
        #Project Description
        self.txtDescription = QPlainTextEdit()
        #Project License
        self.cboLicense = QComboBox()
        self.cboLicense.setFixedWidth(250)
        self.cboLicense.addItem('Apache License 2.0')
        self.cboLicense.addItem('Artistic License/GPL')
        self.cboLicense.addItem('Eclipse Public License 1.0')
        self.cboLicense.addItem('GNU General Public License v2')
        self.cboLicense.addItem('GNU General Public License v3')
        self.cboLicense.addItem('GNU Lesser General Public License')
        self.cboLicense.addItem('MIT License')
        self.cboLicense.addItem('Mozilla Public License 1.1')
        self.cboLicense.addItem('Mozilla Public License 2.0')
        self.cboLicense.addItem('New BSD License')
        self.cboLicense.addItem('Other Open Source')
        self.cboLicense.addItem('Other')
        self.cboLicense.setCurrentIndex(4)
        #Add to Grid
        gbox.addWidget(self.txtName, 0, 1)
        gbox.addLayout(hPlace, 1, 1)
        gbox.addWidget(self.txtDescription, 2, 1)
        gbox.addWidget(self.cboLicense, 3, 1)
        gbox.addLayout(vPlace, 4, 1)
        #Signal
        self.connect(self.btnExamine, SIGNAL('clicked()'), self.load_folder)
        self.connect(self.vbtnExamine, SIGNAL('clicked()'),
            self.load_folder_venv)
        self.connect(self.txtName, SIGNAL('textChanged(const QString&)'),
            lambda: self.emit(SIGNAL("completeChanged()")))

    def isComplete(self):
        name = self.txtName.text().strip()
        place = self.txtPlace.text().strip()
        return (len(name) > 0) and (len(place) > 0)

    def load_folder(self):
        self.txtPlace.setText(QFileDialog.getExistingDirectory(
            self, self.tr("New Project Folder")))
        self.emit(SIGNAL("completeChanged()"))

    def load_folder_venv(self):
        self.vtxtPlace.setText(QFileDialog.getExistingDirectory(
            self, self.tr("Select Virtualenv Folder")))



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import


class ITabItem(object):

    EXTRA_MENU = {}

    def __init__(self):
        self._id = ""    # Should be unique
        self.wasModified = False

        self._parentTab = None

    def get_id(self):
        return self._id

    def set_id(self, id_):
        self._id = id_
        if id_:
            self.newDocument = False

    ID = property(lambda self: self.get_id(), lambda self,
        fileName: self.set_id(fileName))

    def __eq__(self, path):
        """Compares if the path (str) received is equal to the id"""
        return self._id == path

    @classmethod
    def add_extra_menu(cls, menu, lang="py"):
        if not lang in cls.EXTRA_MENU:
            cls.EXTRA_MENU[lang] = []

        cls.EXTRA_MENU[lang].append(menu)



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import
from __future__ import unicode_literals

from PyQt4.QtCore import QThread

from ninja_ide.core import file_manager
from ninja_ide.core import settings
from ninja_ide.dependencies import pep8mod


class Pep8Checker(QThread):

    def __init__(self, editor):
        QThread.__init__(self)
        self._editor = editor
        self._path = ''
        self._encoding = ''
        self.pep8checks = {}

    def check_style(self):
        if not self.isRunning():
            self._path = self._editor.ID
            self._encoding = self._editor.encoding
            self.start()

    def reset(self):
        self.pep8checks = {}

    def run(self):
        self.sleep(1)
        exts = settings.SYNTAX.get('python')['extension']
        file_ext = file_manager.get_file_extension(self._path)
        if file_ext in exts:
            self.reset()
            source = self._editor.get_text()
            tempData = pep8mod.run_check(self._path, source)
            i = 0
            while i < len(tempData):
                lineno = -1
                try:
                    offset = 2 + len(file_ext)
                    startPos = tempData[i].find('.%s:' % file_ext) + offset
                    endPos = tempData[i].find(':', startPos)
                    lineno = int(tempData[i][startPos:endPos]) - 1
                    error = tempData[i][tempData[i].find(
                        ':', endPos + 1) + 2:]
                    line = '\n'.join(
                        [error, tempData[i + 1], tempData[i + 2]])
                except Exception:
                    line = ''
                finally:
                    i += 3
                if line and lineno > -1:
                    if lineno not in self.pep8checks:
                        self.pep8checks[lineno] = [line]
                    else:
                        message = self.pep8checks[lineno]
                        message += [line]
                        self.pep8checks[lineno] = message
        else:
            self.reset()



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

# Based on: https://bitbucket.org/henning/syntaxhighlighter
"""
Partition-based syntax highlighter
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import re

from PyQt4.QtGui import (
    QSyntaxHighlighter,
    QColor, QTextCharFormat, QTextBlockUserData, QFont, QBrush, QTextFormat)

from ninja_ide import resources
from ninja_ide.core import settings


try:
    unicode
except NameError:
    # Python 3
    basestring = unicode = str  # lint:ok


def get_user_data(block):
    user_data = block.userData()
    if user_data is None or not isinstance(user_data, SyntaxUserData):
        user_data = SyntaxUserData()

    return user_data


class SyntaxUserData(QTextBlockUserData):
    """Store the information of the errors, str and comments for each block."""

    def __init__(self, error=False):
        super(SyntaxUserData, self).__init__()
        self.error = error
        self.str_groups = []
        self.comment_start = -1

    def clear_data(self):
        """Clear the data stored for the current block."""
        self.error = False
        self.str_groups = []
        self.comment_start = -1

    def add_str_group(self, start, end):
        """Add a pair of values setting the beggining and end of a string."""
        self.str_groups.append((start, end + 1))

    def comment_start_at(self, pos):
        """Set the position in the line where the comment starts."""
        self.comment_start = pos


class TextCharFormat(QTextCharFormat):
    NAME = QTextFormat.UserProperty + 1


class Format(object):

    __slots__ = ("name", "tcf")

    def __init__(self, name, color=None, bold=None, italic=None,
        base_format=None, background=None):
        self.name = name
        tcf = TextCharFormat()
        if base_format is not None:
            if isinstance(base_format, Format):
                base_format = base_format.tcf
            tcf.merge(base_format)
            tcf.setFont(base_format.font())
        if color is not None:
            if not isinstance(color, QColor):
                color = QColor(color)
            tcf.setForeground(QBrush(color))
        if bold is not None:
            if bold:
                tcf.setFontWeight(QFont.Bold)
            else:
                tcf.setFontWeight(QFont.Normal)
        if italic is not None:
            tcf.setFontItalic(italic)
        if background is not None:
            color = QColor(background)
            tcf.setBackground(color)
        tcf.setProperty(TextCharFormat.NAME, name)
        self.tcf = tcf


class HighlighterError(Exception):
    pass


class Partition(object):
    # every partition maps to a specific state in QSyntaxHighlighter

    __slots__ = ("name", "start", "end", "is_multiline", "search_end")

    def __init__(self, name, start, end, is_multiline=False):
        self.name = name
        self.start = start
        self.end = end
        self.is_multiline = is_multiline
        try:
            self.search_end = re.compile(end, re.M | re.S).search
        except Exception as exc:
            raise HighlighterError("%s: %s %s" % (exc, name, end))


class PartitionScanner(object):
    # The idea to partition the source into different contexts comes
    # from Eclipse.
    # http://wiki.eclipse.org/FAQ_What_is_a_document_partition%3F

    def __init__(self, partitions):
        start_groups = []
        self.partitions = []
        for i, p in enumerate(partitions):
            if isinstance(p, (tuple, list)):
                p = Partition(*p)
            elif isinstance(p, dict):
                p = Partition(**p)
            else:
                assert isinstance(p, Partition), \
                    "Partition expected, got %r" % p
            self.partitions.append(p)
            start_groups.append("(?P<g%s_%s>%s)" % (i, p.name, p.start))
        start_pat = "|".join(start_groups)
        try:
            self.search_start = re.compile(start_pat, re.M | re.S).search
        except Exception as exc:
            raise HighlighterError("%s: %s" % (exc, start_pat))

    def scan(self, current_state, text):
        last_pos = 0
        length = len(text)
        parts = self.partitions
        search_start = self.search_start
        # loop yields (start, end, partition, new_state, is_inside)
        while last_pos < length:
            if current_state == -1:
                found = search_start(text, last_pos)
                if found:
                    start, end = found.span()
                    yield last_pos, start, None, -1, True
                    current_state = found.lastindex - 1
                    p = parts[current_state]
                    yield start, end, p.name, current_state, False
                    last_pos = end
                else:
                    current_state = -1
                    yield last_pos, length, None, -1, True
                    break
            else:
                p = parts[current_state]
                found = p.search_end(text, last_pos)
                if found:
                    start, end = found.span()
                    yield last_pos, start, p.name, current_state, True
                    yield start, end, p.name, current_state, False
                    last_pos = end
                    current_state = -1
                else:
                    yield last_pos, length, p.name, current_state, True
                    break
        if current_state != -1:
            p = parts[current_state]
            if not p.is_multiline:
                current_state = -1
        yield length, length, None, current_state, False


class Token(object):
    __slots__ = ("name", "pattern", "prefix", "suffix")

    def __init__(self, name, pattern, prefix="", suffix=""):
        self.name = name
        if isinstance(pattern, list):
            pattern = "|".join(pattern)
        self.pattern = pattern
        self.prefix = prefix
        self.suffix = suffix


class Scanner(object):
    __slots__ = ("groups", "tokens", "search")

    def __init__(self, tokens):
        self.tokens = []
        self.groups = []
        for t in tokens:
            if isinstance(t, (list, tuple)):
                t = Token(*t)
            elif isinstance(t, dict):
                t = Token(**t)
            else:
                assert isinstance(t, Token), "Token expected, got %r" % t
            gdef = "?P<%s>" % t.name
            if gdef in t.pattern:
                p = t.pattern
            else:
                p = ("(%s%s)" % (gdef, t.pattern))
            p = t.prefix + p + t.suffix
            self.groups.append(p)
            self.tokens.append(t)
        pat = "|".join(self.groups)
        self.search = re.compile(pat).search

    def add_token(self, token):
        self.__clean_highlight()
        tpattern = token[1]
        if tpattern != "":
            t = Token(*token)
            gdef = "?P<%s>" % t.name
            if gdef in t.pattern:
                p = t.pattern
            else:
                p = ("(%s%s)" % (gdef, t.pattern))
            p = t.prefix + p + t.suffix
            self.groups.append(p)
            self.tokens.append(t)
        pat = "|".join(self.groups)
        self.search = re.compile(pat).search

    def __clean_highlight(self):
        for group in self.groups:
            if group.startswith('(?P<highlight_word>'):
                self.groups.remove(group)
                break
        for token in self.tokens:
            if token.name == "highlight_word":
                self.tokens.remove(token)
                break

    def scan(self, s):
        search = self.search
        #length = len(s)
        last_pos = 0
        # loop yields (token, start_pos, end_pos)
        while 1:
            found = search(s, last_pos)
            if found:
                lg = found.lastgroup
                start, last_pos = found.span(lg)
                yield lg, start, last_pos
            else:
                break


class SyntaxHighlighter(QSyntaxHighlighter):

    def __init__(self, parent, partition_scanner, scanner, formats,
        errors=None, pep8=None, migration=None):
        """
        :param parent: QDocument or QTextEdit/QPlainTextEdit instance
        'partition_scanner:
            PartitionScanner instance
        :param scanner:
            dictionary of token scanners for each partition
            The key is the name of the partition, the value is a Scanner
                instance
            The default scanner has the key None
        :formats:
            list of tuples consisting of a name and a format definition
            The name is the name of a partition or token
        """
        super(SyntaxHighlighter, self).__init__(parent)
        self.errors = errors
        self.pep8 = pep8
        self.migration = migration
        self._old_search = None

        if isinstance(partition_scanner, (list, tuple)):
            partition_scanner = PartitionScanner(partition_scanner)
        else:
            assert isinstance(partition_scanner, PartitionScanner), \
                "PartitionScanner expected, got %r" % partition_scanner
        self.partition_scanner = partition_scanner

        self.scanner = scanner
        for inside_part, inside_scanner in list(scanner.items()):
            if isinstance(inside_scanner, (list, tuple)):
                inside_scanner = Scanner(inside_scanner)
                self.scanner[inside_part] = inside_scanner
            else:
                assert isinstance(inside_scanner, Scanner), \
                    "Scanner expected"

        self.formats = {}
        for f in formats:
            if isinstance(f, tuple):
                fname, f = f
            else:
                assert isinstance(f, (Format, dict)), \
                    "Format expected, got %r" % f
            if isinstance(f, basestring):
                f = (f,)  # only color specified
            if isinstance(f, (tuple, list)):
                f = Format(*((fname,) + f))
            elif isinstance(f, dict):
                f = Format(**dict(name=fname, **f))
            else:
                assert isinstance(f, Format), "Format expected, got %r" % f
            f.tcf.setFontFamily(parent.defaultFont().family())
            self.formats[f.name] = f.tcf

        # reduce name look-ups for better speed
        scan_inside = {}
        for inside_part, inside_scanner in list(self.scanner.items()):
            scan_inside[inside_part] = inside_scanner.scan
        self.get_scanner = scan_inside.get
        self.scan_partitions = partition_scanner.scan
        self.get_format = self.formats.get

    def __apply_proper_style(self, char_format, color):
        if settings.UNDERLINE_NOT_BACKGROUND:
            char_format.setUnderlineColor(color)
            char_format.setUnderlineStyle(QTextCharFormat.WaveUnderline)
        else:
            color.setAlpha(60)
            char_format.setBackground(color)

    def __highlight_pep8(self, char_format, user_data=None):
        """Highlight the lines with pep8 errors."""
        user_data.error = True
        color = QColor(
            resources.CUSTOM_SCHEME.get('pep8-underline',
            resources.COLOR_SCHEME['pep8-underline']))
        self.__apply_proper_style(char_format, color)
        return char_format

    def __highlight_lint(self, char_format, user_data):
        """Highlight the lines with lint errors."""
        user_data.error = True
        color = QColor(
            resources.CUSTOM_SCHEME.get('error-underline',
            resources.COLOR_SCHEME['error-underline']))
        self.__apply_proper_style(char_format, color)
        return char_format

    def __highlight_migration(self, char_format, user_data):
        """Highlight the lines with lint errors."""
        user_data.error = True
        color = QColor(
            resources.CUSTOM_SCHEME.get('migration-underline',
            resources.COLOR_SCHEME['migration-underline']))
        self.__apply_proper_style(char_format, color)
        return char_format

    def __remove_error_highlight(self, char_format, user_data):
        user_data.error = False
        char_format.setUnderlineStyle(QTextCharFormat.NoUnderline)
        char_format.clearBackground()
        return char_format

    def highlightBlock(self, text):
        """automatically called by Qt"""
        text += "\n"
        previous_state = self.previousBlockState()
        new_state = previous_state
        # User data and errors
        block = self.currentBlock()
        user_data = get_user_data(block)
        user_data.clear_data()
        valid_error_line, highlight_errors = self.get_error_highlighter(block)
        # speed-up name-lookups
        get_format = self.get_format
        set_format = self.setFormat
        get_scanner = self.get_scanner

        for start, end, partition, new_state, is_inside in \
          self.scan_partitions(previous_state, text):
            f = get_format(partition, None)
            if f:
                f = highlight_errors(f, user_data)
                set_format(start, end - start, f)
            elif valid_error_line:
                f = TextCharFormat()
                f = highlight_errors(f, user_data)
                set_format(start, end - start, f)
            if is_inside:
                scan = get_scanner(partition)
                if scan:
                    for token, token_pos, token_end in scan(text[start:end]):
                        f = get_format(token)
                        if f:
                            f = highlight_errors(f, user_data)
                            set_format(start + token_pos,
                                token_end - token_pos, f)
            if partition and partition == "string":
                user_data.add_str_group(start, end)

        block.setUserData(user_data)
        self.setCurrentBlockState(new_state)

    def get_error_highlighter(self, block):
        block_number = block.blockNumber()
        highlight_errors = self.__remove_error_highlight
        valid_error_line = False
        if self.errors and (block_number in self.errors.errorsSummary):
            highlight_errors = self.__highlight_lint
            valid_error_line = True
        elif self.pep8 and (block_number in self.pep8.pep8checks):
            highlight_errors = self.__highlight_pep8
            valid_error_line = True
        elif self.migration and (
             block_number in self.migration.migration_data):
            highlight_errors = self.__highlight_migration
            valid_error_line = True
        return valid_error_line, highlight_errors

    def set_selected_word(self, word, partial=False):
        """Set the word to highlight."""
        hl_worthy = len(word) > 2
        if hl_worthy:
            suffix = "(?![A-Za-z_\d])"
            prefix = "(?<![A-Za-z_\d])"
            word = re.escape(word)
            if not partial:
                word = "%s%s%s" % (prefix, word, suffix)
            self.scanner[None].add_token(('highlight_word', word))
        else:
            self.scanner[None].add_token(('highlight_word', ""))
        lines = []
        pat_find = re.compile(word)
        document = self.document()
        for lineno, text in enumerate(document.toPlainText().splitlines()):
            if hl_worthy and pat_find.search(text):
                lines.append(lineno)
            elif self._old_search and self._old_search.search(text):
                lines.append(lineno)
        # Ask perrito if i don't know what the next line does:
        self._old_search = hl_worthy and pat_find
        self.rehighlight_lines(lines)

    def _rehighlight_lines(self, lines):
        """If the document is valid, highlight the list of lines received."""
        if self.document() is None:
            return
        for line in lines:
            block = self.document().findBlockByNumber(line)
            self.document().markContentsDirty(block.position(),
                block.position() + block.length())
            self.rehighlightBlock(block)

    def _get_errors_lines(self):
        """Return the number of lines that contains errors to highlight."""
        errors_lines = []
        block = self.document().begin()
        while block.isValid():
            user_data = get_user_data(block)
            if user_data.error:
                errors_lines.append(block.blockNumber())
            block = block.next()
        return errors_lines

    def rehighlight_lines(self, lines, errors=True):
        """Rehighlight the lines for errors or selected words."""
        if errors:
            errors_lines = self._get_errors_lines()
            refresh_lines = set(lines + errors_lines)
        else:
            refresh_lines = set(lines)
        self._rehighlight_lines(refresh_lines)


def _create_scheme():
    scheme = {
      "syntax_comment": dict(color=resources.CUSTOM_SCHEME.get(
          "comment", resources.COLOR_SCHEME["comment"]), italic=True),
      "syntax_string": resources.CUSTOM_SCHEME.get(
          "string", resources.COLOR_SCHEME["string"]),
      "syntax_builtin": resources.CUSTOM_SCHEME.get(
          "extras", resources.COLOR_SCHEME["extras"]),
      "syntax_keyword": (resources.CUSTOM_SCHEME.get(
          "keyword", resources.COLOR_SCHEME["keyword"]), True),
      "syntax_definition": (resources.CUSTOM_SCHEME.get(
          "definition", resources.COLOR_SCHEME["definition"]), True),
      "syntax_braces": resources.CUSTOM_SCHEME.get(
          "brace", resources.COLOR_SCHEME["brace"]),
      "syntax_number": resources.CUSTOM_SCHEME.get(
          "numbers", resources.COLOR_SCHEME["numbers"]),
      "syntax_proper_object": resources.CUSTOM_SCHEME.get(
          "properObject", resources.COLOR_SCHEME["properObject"]),
      "syntax_operators": resources.CUSTOM_SCHEME.get(
          "operator", resources.COLOR_SCHEME["operator"]),
      "syntax_highlight_word": dict(color=resources.CUSTOM_SCHEME.get(
          "selected-word", resources.COLOR_SCHEME["selected-word"]),
          background=resources.CUSTOM_SCHEME.get(
          "selected-word-background",
          resources.COLOR_SCHEME["selected-word-background"])),
      "syntax_pending": resources.COLOR_SCHEME["pending"],
    }

    return scheme


def load_syntax(syntax):
    context = _create_scheme() or {}

    partition_scanner = PartitionScanner(syntax.get("partitions", []))

    scanners = {}
    for part_name, part_scanner in list(syntax.get("scanner", {}).items()):
        scanners[part_name] = Scanner(part_scanner)

    formats = []
    for fname, fstyle in list(syntax.get("formats", {}).items()):
        if isinstance(fstyle, basestring):
            if fstyle.startswith("%(") and fstyle.endswith(")s"):
                key = fstyle[2:-2]
                fstyle = context[key]
            else:
                fstyle = fstyle % context
        formats.append((fname, fstyle))

    return partition_scanner, scanners, formats
